import sys
import os
import time
import threading
import logging
import pdb
import Queue
import saga
import uuid
import traceback
import urlparse

import bigjob
import pilot
from bigjob import logger
from pilot.api import ComputeDataService, ComputeUnit, State
from pilot.impl.pilotdata_manager import PilotData, DataUnit
#from pilot.coordination.advert import AdvertCoordinationAdaptor as CoordinationAdaptor
from pilot.coordination.nocoord import NoCoordinationAdaptor as CoordinationAdaptor
""" Loaded Module determines scheduler:
    
    bigdata.scheduler.data_compute_scheduler - selects random locations for PD and WUs
    bigdata.scheduler.data_compute_affinity_scheduler - considers affinity descriptions
    
"""

from pilot.scheduler.data_compute_affinity_scheduler import Scheduler

class ComputeDataService(ComputeDataService):
    """ TROY WorkDataService.
    
        The WorkDataService is the application's interface to submit 
        ComputeUnits and PilotData/DataUnit to the Pilot-Manager 
        in the P* Model.
    """    
    CDS_ID_PREFIX="cds-"  


    def __init__(self, cds_url=None):
        """ Create a Work Data Service object.

            Keyword arguments:
            cds_url -- Reconnect to an existing WDS (optional).
        """
        # Pilot Data
        self.data_units={}
        self.pilot_data_services=[]
        
        # Pilot Job
        self.compute_units={}
        self.pilot_job_services=[]
            
        if cds_url == None:
            self.id=self.CDS_ID_PREFIX + str(uuid.uuid1())
            application_url = CoordinationAdaptor.get_base_url(pilot.application_id)
            self.url = CoordinationAdaptor.add_cds(application_url, self)
            
        else:
            self.id = self.__get_cds_id(cds_url)
            self.url = cds_url
           
        # Background Thread for scheduling
        self.scheduler = Scheduler()
        self.cu_queue = Queue.Queue()
        self.du_queue = Queue.Queue()
        self.stop=threading.Event()
        self.scheduler_thread=threading.Thread(target=self._scheduler_thread)
        self.scheduler_thread.start()

    def __get_cds_id(self, cds_url):
        start = cds_url.index(self.CDS_ID_PREFIX)
        end =cds_url.index("/", start)
        return cds_url[start:end]


    ###########################################################################
    # Pilot Job
    
    def add_pilot_compute_service(self, pjs):
        """ Add a PilotJobService to this CDS.

            Keyword arguments:
            pilotjob_services -- The PilotJob Service(s) to which this 
                                 Work Unit Service will connect.

            Return:
            Result
        """
        self.pilot_job_services.append(pjs)
        CoordinationAdaptor.update_cds(self.url, self)

    def remove_pilot_compute_service(self, pjs):
        """ Remove a PilotJobService from this CDS.

            Note that it won't cancel the PilotJobService, it will just no
            longer be connected to this WUS.

            Keyword arguments:
            pilotjob_services -- The PilotJob Service(s) to remove from this
                                 Work Unit Service. 

            Return:
            Result
        """
        self.pilot_job_services.remove(pjs)
        CoordinationAdaptor.update_cds(self.url, self)

    def submit_compute_unit(self, compute_unit_description):
        """ Submit a WU to this Work Unit Service.

            Keyword argument:
            cud -- The ComputeUnitDescription from the application

            Return:
            ComputeUnit object
        """
        cu = ComputeUnit(compute_unit_description, self)
        self.compute_units[cu.id]=cu
        self.cu_queue.put(cu)
        CoordinationAdaptor.update_cds(self.url, self)
        return cu
    
    
    ###########################################################################
    # Pilot Data     
    def add_pilot_data_service(self, pds):
        """ Add a PilotDataService 

            Keyword arguments:
            pds -- The PilotDataService to add.

            Return:
            None
        """
        self.pilot_data_services.append(pds)
        CoordinationAdaptor.update_cds(self.url, self)
    
    def remove_pilot_data_service(self, pds):

        """ Remove a PilotDataService 
            
            Keyword arguments:
            pds -- The PilotDataService to remove 
            
            Return:
            None
        """
        self.pilot_data_services.remove(pds)
        CoordinationAdaptor.update_cds(self.url, self)
    
    
    def list_pilot_compute(self):
        """ List all pilot compute of CDS """
        return self.pilot_job_service
    
    
    def list_pilot_data(self):
        """ List all pilot data of CDS """
        return self.pilot_data_services
    
    
    def list_data_units(self):
        """ List all DUs of CDS """
        return self.data_units.items()
    
    
    def get_data_unit(self, du_id):
        if self.data_units.has_key(du_id):
            return self.data_units[du_id]
        return None
    
    
    def submit_data_unit(self, data_unit_description):
        """ creates a data unit object and binds it to a physical resource (a pilotdata) """
        du = DataUnit(pilot_data_service=self, 
                      data_unit_description=data_unit_description)
        self.data_units[du.id]=du
        self.du_queue.put(du)
        # queue currently not persisted
        CoordinationAdaptor.update_cds(self.url, self)
        return du
    
    
    def cancel(self):
        """ Cancel the PDS. 
            All associated PD objects are deleted and removed from the associated pilot stores.            
            
            Keyword arguments:
            None

            Return:
            None
        """
        # terminate background thread
        self.stop.set()
        CoordinationAdaptor.delete_cds(self.url)
   
    def wait(self):
        """ Waits for CUs and DUs
            Return if all du's are running 
                   AND
                   cu's are done
            
        """
        logger.debug("### START WAIT ###")
        self.cu_queue.join()
        logger.debug("CU queue empty")        
        self.du_queue.join()
        logger.debug("DU queue empty")        

        for i in self.data_units.values():
            i.wait()
        logger.debug("DUs done")        
            
        for i in self.compute_units.values():
            i.wait()     
        logger.debug("CUs done")        
               
        logger.debug("### END WAIT ###")

        #[pc.wait() for i in self.pilot_job_services for pc in i.list_pilots()]
        #[pd.wait() for i in self.pilot_data_services for pd in i.list_pilots()]
                
        
    def get_state(self):
        return self.state
    
    
    def get_id(self):
        return str(self.id)
   
    ###########################################################################
    # Internal Scheduling
    def __update_scheduler_resources(self):
        logging.debug("__update_scheduler_resources")        
        pd = [s for i in self.pilot_data_services for s in i.list_pilots()]
        self.scheduler.set_pilot_data(pd)
        pj = [p for i in self.pilot_job_services for p in i.list_pilots()]
        logging.debug("Pilot-Jobs: " + str(pj))
        self.scheduler.set_pilot_jobs(pj)
    
    def _schedule_du(self, du):
        """ Schedule DU to a suitable pilot data
        
            Currently one level of scheduling is used:
                1.) Add all resources managed by the contained PDS 
                2.) Select one resource
        """ 
        logging.debug("Schedule PD")
        self.__update_scheduler_resources()
        selected_pilot_data = self.scheduler.schedule_pilot_data(du.data_unit_description)
        return selected_pilot_data 
    
    def _schedule_cu(self, cu):
        logging.debug("Schedule CU")
        self.__update_scheduler_resources()
        selected_pilot_job = self.scheduler.schedule_pilot_job(cu.compute_unit_description)
        return selected_pilot_job
    
    def _scheduler_thread(self):
        while True and self.stop.isSet()==False:            
            try:
                logging.debug("Scheduler Thread: " + str(self.__class__) + " Pilot Data")
                du = self.du_queue.get(True, 1)  
                # check whether this is a real du object  
                if isinstance(du, DataUnit):
                    pd=self._schedule_du(du)                
                    if(pd!=None):                        
                        logging.debug("Initiate Transfer to PD.")
                        du.add_pilot_data(pd)
                        logging.debug("Transfer to PD finished.")
                        du.update_state(State.Running) 
                        self.du_queue.task_done()                   
                    else:
                        self.du_queue.task_done() 
                        self.du_queue.put(du)
            except Queue.Empty:
                pass
                    
            try:    
                logging.debug("Scheduler Thread: " + str(self.__class__) + " Pilot Job")
                cu = self.cu_queue.get(True, 1)                
                if isinstance(cu, ComputeUnit):                    
                    pj=self._schedule_cu(cu) 
                    if pj !=None:
                        cu = self.__expand_working_directory(cu, pj)                        
                        pj._submit_cu(cu)           
                        self.cu_queue.task_done()         
                    else:
                        self.cu_queue.task_done() 
                        self.cu_queue.put(du)
            except Queue.Empty:
                pass
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                logger.error("*** print_tb:")
                traceback.print_tb(exc_traceback, limit=1, file=sys.stderr)
                logger.error("*** print_exception:")
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stderr)
            time.sleep(5)        

        logging.debug("Re-Scheduler terminated")
    
    
    def __expand_working_directory(self, compute_unit, pilot_job):
        """ Expand pilotdata:// url specified in the compute_unit_description 
            to a local url on the machine of the PJ
            
            pilotdata://localhost/434bfc5c-23fd-11e1-a43f-00264a13ca4c
            
            to
            
           /tmp/pilotstore//434bfc5c-23fd-11e1-a43f-00264a13ca4c on machine running pilot_job        
        """ 
        if compute_unit.compute_unit_description.has_key("working_directory"):
            working_directory=compute_unit.compute_unit_description["working_directory"]       
            if working_directory.find(DataUnit.DU_ID_PREFIX)!=-1:
                pilot_data_url = working_directory
                pj_description = pilot_job.pilot_compute_description
                pj_dc_affinity = pj_description["affinity_datacenter_label"]
                pj_machine_affinity = pj_description["affinity_machine_label"]
                pd = [s for i in self.pilot_data_services for s in i.list_pilots()]
                
                # find all pilot stores with the same affinity
                candidate_pd = []
                for i in pd:
                    pd_description = i.pilot_data_description
                    pd_dc_affinity = pd_description["affinity_datacenter_label"]
                    pd_machine_affinity = pd_description["affinity_machine_label"]
                    if pd_dc_affinity == pj_dc_affinity and pd_machine_affinity == pj_machine_affinity:
                        candidate_pd.append(i)
                    
                # check whether required pilot_data is part of pilot_data
                target_pd = None  
                target_du = None  
                for pd in candidate_pd:
                    for du in pd.list_data_units():
                        logger.debug("DU URL:%s"%(du.url))
                        if du.url == pilot_data_url:
                            logging.debug("Found PD %s at %s"%(du.url, pd.service_url))
                            target_pd = pd 
                            target_du = du
                            break
                if target_du == None:
                    self.__stage_du_to_pj(pilot_data_url, pilot_job)
                
                if target_pd!=None:
                    pd_url = target_pd.url_for_du(target_du)
                    components = urlparse.urlparse(pd_url)
                    compute_unit.compute_unit_description["working_directory"] = components.path
                    compute_unit._update_compute_unit_description(compute_unit.compute_unit_description)
                    logging.debug("__expand_working_directory %s: Set working directory to %s"%(pilot_data_url, compute_unit.compute_unit_description["working_directory"]))
                    return compute_unit
         
        return compute_unit
            
            
    def __stage_du_to_pj(self, pilotdata, pilotjob):
        """
            stage required files to machine of pilot job
        """
        pass
    
    def __find_pd_at_pj_resource(self, pilotjob):
        pass
   
    
    
class ComputeUnit(ComputeUnit):
    """ ComputeUnit - Wrapper for BigJob subjob """
    CU_ID_PREFIX="cu-"  

    def __init__(self, compute_unit_description, work_data_service):
        self.id = self.CU_ID_PREFIX + str(uuid.uuid1())
        self.url = work_data_service.url + "/" + self.id
        self.state = State.New       
        self.subjob = None # reference to BigJob Subjob 
        self.compute_unit_description = compute_unit_description # WU Description
        self.subjob_description = self.__translate_cu_sj_description(compute_unit_description)
        logger.debug("Created CU: %s"%self.url)     
                
    def get_id(self):
        return self.id
    
    
    def get_state(self):
        if self.subjob != None:
            return self.subjob.get_state()
        return self.state
    
    
    def wait(self):
        """ Wait until in Done state 
            (or Failed state)
        """
        while True:
            state = self.get_state()
            logger.debug("Compute Unit - State: %s"%state)            
            if state==State.Done or state==State.Failed:
                break
            time.sleep(2)
            
        #logger.debug("### END CU wait")

    
    def cancel(self):
        if self.subjob != None:
            return self.subjob.cancel()
        return None

    
    def _update_compute_unit_description(self, compute_unit_description):
        self.compute_unit_description = compute_unit_description # WU Description
        self.subjob_description = self.__translate_cu_sj_description(compute_unit_description)
        
    # INTERNAL
    def __translate_cu_sj_description(self, compute_unit_description):
        jd = saga.job.description()
        if compute_unit_description.has_key("executable"): 
            jd.executable = compute_unit_description["executable"]
        jd.spmd_variation = "single"
        if compute_unit_description.has_key("arguments"): 
            jd.arguments = compute_unit_description["arguments"]
        
        if compute_unit_description.has_key("number_of_processes"):
            jd.number_of_processes=str(compute_unit_description["number_of_processes"])
        else:
            jd.number_of_processes="1"
        
        if compute_unit_description.has_key("working_directory"): 
            jd.working_directory = compute_unit_description["working_directory"]
        if compute_unit_description.has_key("output"): 
            jd.output =  compute_unit_description["output"]
        if compute_unit_description.has_key("error"): 
            jd.error = compute_unit_description["error"]
        return jd
        



application_id = "bigjob"


from pilot.impl.pilotcompute_manager import PilotComputeService as MyPilotComputeService
from pilot.impl.pilotdata_manager import PilotDataService as MyPilotDataService
from pilot.impl.pilotdata_manager import DataUnit as MyDataUnit
from pilot.impl.pilot_manager import ComputeDataService as MyComputeDataService
from pilot.api import State as MyState

# define external-facing API

class PilotComputeService(MyPilotComputeService):
    pass


class PilotDataService(MyPilotDataService):
    pass


class DataUnit(MyDataUnit):
    pass


class ComputeDataService(MyComputeDataService):
    pass


class State(MyState):
    pass





#!/usr/bin/env python

import textwrap
import re
import pdb

from bigjob import logger
import bigjob

try:
    import saga
except:
    logger.warn("pbs-ssh://<hostname> plugin not compatible with SAGA Bliss. Use pbs+ssh://<hostname>")

import os

class pbsssh:
    """Constructor"""
    def __init__(self, bootstrap_script, launch_method, queue, project, lrms_saga_url, walltime, number_nodes, processes_per_node, userproxy, working_directory=None, bj_working_directory=None):
        self.job_id = ""
        self.lrms_saga_url = lrms_saga_url
        if launch_method == "ssh":
            self.lrms_saga_url.scheme="ssh"
        else:
            self.lrms_saga_url.scheme="gsissh"
        self.userproxy = userproxy
        self.working_directory = ""
        if working_directory == None:
            self.working_directory = ""
        else:
            self.working_directory = working_directory
        if bj_working_directory==None:
            bj_working_directory=self.working_directory
        ### convert walltime in minutes to PBS representation of time ###
        walltime_pbs="1:00:00"
        if walltime!=None and walltime!="":    
            hrs=int(walltime)/60 
            minu=int(walltime)%60 
            walltime_pbs=""+str(hrs)+":"+str(minu)+":00"

        if int(number_nodes)%int(processes_per_node) == 0:
            nodes = int(number_nodes)/int(processes_per_node)
        else:
            nodes = (int(number_nodes)/int(processes_per_node)) + 1    
        
        ppn = processes_per_node
        self.bootstrap_script = textwrap.dedent("""import sys
import os
import urllib
import sys
import time
import textwrap

qsub_file_name="bigjob_pbs_ssh"

qsub_file = open(qsub_file_name, "w")
if ( %s == 1 ):
    qsub_file.write("#PBS -l size=%s")
else:
    qsub_file.write("#PBS -l nodes=%s:ppn=%s")
qsub_file.write("\\n")
if ( "%s" != "None" ):
    qsub_file.write("#PBS -q %s")
    qsub_file.write("\\n")
if (  "%s" != "None" ):
    qsub_file.write("#PBS -A %s")
    qsub_file.write("\\n")    
qsub_file.write("#PBS -l walltime=%s")
qsub_file.write("\\n")
qsub_file.write("#PBS -o %s/stdout-bigjob_agent.txt")
qsub_file.write("\\n")
qsub_file.write("#PBS -e %s/stderr-bigjob_agent.txt")
qsub_file.write("\\n")
qsub_file.write("cd %s")
qsub_file.write("\\n")
qsub_file.write("python -c XX" + textwrap.dedent(\"\"%s\"\") + "XX")
qsub_file.close()
os.system( "qsub  " + qsub_file_name)
""") % (str(ppn),str(nodes),str(nodes),str(ppn),str(queue),str(queue),str(project),str(project),str(walltime_pbs), bj_working_directory, bj_working_directory, str(self.working_directory), bootstrap_script)
        ### escaping characters
        self.bootstrap_script = self.bootstrap_script.replace("\"","\\\"")
        self.bootstrap_script = self.bootstrap_script.replace("\\\\","\\\\\\\\\\")
        self.bootstrap_script = self.bootstrap_script.replace("XX","\\\\\\\"")
        self.bootstrap_script = "\"" + self.bootstrap_script+ "\""
        logger.debug(self.bootstrap_script)


    def run(self):
        jd = saga.job.description()
        jd.arguments = ["-c", self.bootstrap_script]
        jd.executable = "python"
        jd.working_directory =  self.working_directory
        jd.set_attribute("Interactive", "True")
        # Submit job
        js = None
        if self.userproxy != None and self.userproxy != '':
            s = saga.session()
            os.environ["X509_USER_PROXY"]=self.userproxy
            ctx = saga.context("x509")
            ctx.set_attribute ("UserProxy", self.userproxy)
            s.add_context(ctx)
            print "use proxy: " + self.userproxy
            js = saga.job.service(s, self.lrms_saga_url)
        else:
            print "use standard proxy"
            js = saga.job.service(self.lrms_saga_url)
        pbssshjob = js.create_job(jd)
        print "Submit pilot job to: " + str(self.lrms_saga_url)
        pbssshjob.run()
        pbssshjob.wait()
        outstr = pbssshjob.get_stdout().read()
        errstr = pbssshjob.get_stderr().read()
        self.job_id=(outstr).split(".")[0]
        logger.debug("PBS JobID: " + str(self.job_id))
        if self.job_id==None or self.job_id=="":
            raise Exception("BigJob submission via pbs-ssh:// failed: %s %s" % (outstr,errstr))


    def get_state(self):
        jd = saga.job.description()
        jd.set_attribute("Executable", "qstat")
        jd.set_attribute("Interactive", "True")
        jd.set_vector_attribute("Arguments", [self.job_id])
        # connect to the local job service
        js = saga.job.service(self.lrms_saga_url);
        # submit the job
        job = js.create_job(jd)
        job.run()
        # wait for the job to complete
        job.wait(-1)
        # print the job's output
        output = job.get_stdout()
        output.readline()
        output.readline()
        state=output.readline()
        state=(re.sub(r'\s+',' ',state)).split(' ')[4]
        if state == "R":
            state = "Running"
        elif state == "C" or state == "E":
            state = "Done"
        else:
            state = "Unknown"
        return state 

    def cancel(self):
        jd = saga.job.description()
        jd.set_attribute("Executable", "qdel")
        jd.set_attribute("Interactive", "True")
        jd.set_vector_attribute("Arguments", [self.job_id])
        # connect to the local job service
        js = saga.job.service(self.lrms_saga_url);
        # submit the job
        job = js.create_job(jd)
        job.run()
        # wait for the job to complete
        job.wait(-1)



#!/usr/bin/env python

"""
Dynamic BigJob (ManyJob) Example

This Module is used to launch a set of bigjobs.

DON'T EDIT THIS FILE (UNLESS THERE IS A BUG)
    
THIS FILE SHOULD NOT BE COMMITTED TO SVN WITH USE-SPECIFIC PATHS!

"""

import sys
import getopt
import time
import pdb
import os
import traceback
import logging


# BigJob implementation can be swapped here by importing another implementation,
# e.g. condor, cloud, azure
import sys

sys.path.insert(0, os.getcwd() + "/../")

from bigjob import bigjob, subjob, description
from bigjob_dynamic.many_job import *


""" This variable defines the coordination system that is used by BigJob
    e.g. 
        advert://localhost (SAGA/Advert SQLITE)
        advert://advert.cct.lsu.edu:8080 (SAGA/Advert POSTGRESQL)
        redis://localhost:6379 (Redis at localhost)
        tcp://localhost (ZMQ)
"""
COORDINATION_URL = "advert://localhost"
#COORDINATION_URL = "advert://advert.cct.lsu.edu:8080"
    
NUMBER_JOBS=8

def has_finished(state):
        state = state.lower()
        if state=="done" or state=="failed" or state=="canceled":
            return True
        else:
            return False

def main():
    try:
        print "ManyJob load test with " + str(NUMBER_JOBS) + " jobs."
        starttime=time.time()

        """ submit via mj abstraction
         
         resource_list.append( {"resource_url" : "gram://eric1.loni.org/jobmanager-pbs", "processes_per_node":"4",
                               "number_of_processes" : "4", "allocation" : None, "queue" : "workq", 
                               "working_directory": (os.getcwd() + "/agent"), "walltime":10 })

        """
        resource_list = []
        resource_dictionary = {"resource_url" : "fork://localhost/", "number_of_processes" : "32", 
                               "processes_per_node":"1", "allocation" : None, "queue" : None, 
                               "working_directory": (os.getcwd() + "/agent"), "walltime":3600 }
        resource_list.append(resource_dictionary)
        
        
        #Flags for controlling dynamic BigJob
        add_additional_resources=True
        remove_additional_resources=False
        

        print "Create Dynamic BigJob Service "
        mjs = many_job_service(resource_list, COORDINATION_URL)
        
        jobs = []
        job_start_times = {}
        job_states = {}
        cwd = os.getcwd()
        for i in range(0, NUMBER_JOBS):
            # create job description
            jd = description()
            jd.executable = "/bin/date"
            jd.number_of_processes = "1"
            jd.spmd_variation = "single"
            jd.arguments = [""]
            jd.working_directory = os.getcwd();
            jd.output =  "stdout-" + str(i) + ".txt"
            jd.error = "stderr-" + str(i) + ".txt"
            subjob = mjs.create_job(jd)
            subjob.run()
            print "Submited sub-job " + "%d"%i + "."
            jobs.append(subjob)
            job_start_times[subjob]=time.time()
            job_states[subjob] = subjob.get_state()
        print "************************ All Jobs submitted ************************"
        while 1: 
            finish_counter=0
            result_map = {}
            for i in range(0, NUMBER_JOBS):
                old_state = job_states[jobs[i]]
                state = jobs[i].get_state()
                if result_map.has_key(state) == False:
                    result_map[state]=0
                result_map[state] = result_map[state]+1
                #print "counter: " + str(i) + " job: " + str(jobs[i]) + " state: " + state
                if old_state != state:
                    print "Job " + str(jobs[i]) + " changed from: " + old_state + " to " + state
                if old_state != state and has_finished(state)==True:
                    print "Job: " + str(jobs[i]) + " Runtime: " + str(time.time()-job_start_times[jobs[i]]) + " s."
                if has_finished(state)==True:
                    finish_counter = finish_counter + 1
                job_states[jobs[i]]=state
                
            # Dynamic BigJob add resources at runtime
            # if more than 30 s - add additional resource
            if time.time()-starttime > 10 and add_additional_resources==True:
                print "***add additional resources***"
                mjs.add_resource(resource_dictionary)
                add_additional_resources=False  
                
            # remove resources from dynamic bigjob
            if (time.time()-starttime > 15 and remove_additional_resources==True):
                bj_list = mjs.get_resources()
                if len(bj_list)>0:
                    print "***remove resources: " + str(bj_list[0])
                    mjs.remove_resource(bj_list[0])
                remove_additional_resources=False
                
            print "Current states: " + str(result_map) 
            time.sleep(5)
            if finish_counter == NUMBER_JOBS:
                break

        mjs.cancel()
        runtime = time.time()-starttime
        print "Runtime: " + str(runtime) + " s; Runtime per Job: " + str(runtime/NUMBER_JOBS)
    except:
        traceback.print_exc(file=sys.stdout)
        try:
            mjs.cancel()
        except:
            pass

""" Test Job Submission via ManyJob abstraction """
if __name__ == "__main__":
    main()



""" 
This file contains the API for the PilotJob Framework.
It is not supposed to be used directly by applications. 
"""

class State(object):
    Unknown = "Unknown"
    New = "New"
    Running = "Running"
    Done = "Done"
    Canceled = "Canceled"
    Failed = "Failed"
    Pending = "Pending"
    


class PilotComputeDescription(dict):
    """  PilotComputeDescription.
        
        A PilotComputeDescription is a based on the SAGA Job Description.

        The PilotComputeDescription is used by the application to specify 
        what kind of PilotJobs it requires.
    """

    # Class members
    __slots__ = (
        # Pilot / Agent description
        'executable',
        'arguments',
        'cleanup',
        'environment',
        'interactive',
        'contact',
        'project',
        'start_time',
        'working_directory',
        # I/O
        'input',
        'error',
        'output',
        'file_transfer',
        # Parallelism
        'number_of_processes',      # Total number of processes to start
        'processes_per_host',       # Nr of processes per host
        'threads_per_process',      # Nr of threads to start per process
        'total_core_count',         # total number of cores requested
        'spmd_variation',           # Type and startup mechanism
        
        # Requirements
        'candidate_hosts',
        'cpu_architecture',
        'total_physical_memory', 
        'operating_system_type',
        'total_cpu_time',
        'wall_time_limit',
        'queue',
        
        # Affinity
        'affinity_datacenter_label',    # pilot jobs sharing the same label are located in the same data center          
        'affinity_machine_label',       # pilot jobs sharing the same label are located on the same machine
    )

    def __init__(self):
        pass
    
    
    def __setattr__(self, attr, value):
        self[attr]=value
        
    
    def __getattr__(self, attr):
        return self[attr]
    
    

class PilotCompute(object):
    """ PilotCompute (PilotJob). 
    
        This is the object that is returned by the PilotComputeService when a 
        new PilotCompute is created based on a PilotComputeDescription.

        The PilotCompute object can be used by the application to keep track 
        of PilotComputes that are active.
        
        A PilotCompute has state, can be queried, can be cancelled and be 
        re-initialized.
    """

    # Class members
    __slots__ = (
        'id',           # Reference to this PJ
        'description',  # Description of PilotCompute
        'context',      # SAGA context
        'rm',           # Resource Manager URL
        'pj_type',      # Type of TROY backend
        'state',        # State of the PilotCompute
        'state_detail', # Adaptor specific state of the PilotCompute
        'callback',     # Callback object
        'wall_time_left'      # Remaining wallclock time left
    )

    def cancel(self):        
        """ Remove the PilotCompute from the PilotCompute Service.

            Keyword arguments:
            None
        """
        pass

    def reinitialize(self, pilotjob_description):        
        """ Re-Initialize the PilotCompute to the (new) PilotComputeDescription.
        
            Keyword arguments:
            pilotjob_description -- A PilotComputeDescription
        """
        pass

    def set_callback(self, member, cb):
        """ Set a callback function for a member.

            Keyword arguments:
            member -- The member to set the callback for (state / state_detail).
            cb -- The callback object to call.
        """
        pass

    def unset_callback(self, member):
        """ Unset a callback function from a member

            Keyword arguments:
            member -- The member to unset the callback from.
        """
        pass
    
    def get_state(self):
        pass
#
#  Pilot ComputeService
#
class PilotComputeService(object):
    """  PilotComputeService.
    
        The PilotComputeService is responsible for creating and managing 
        the PilotComputes.

        It is the application's interface to the Pilot-Manager in the 
        P* Model.
        
    """

    # Class members
    __slots__ = (
        'id',           # Reference to this PJS
        'state',       # Status of the PJS
        'pilot_jobs'    # List of PJs under this PJS
    )

    def __init__(self, pjs_id=None):
        """ Create a PilotComputeService object.

            Keyword arguments:
            pjs_id -- Don't create a new, but connect to an existing (optional)
        """
        pass

    def create_pilot(self, rm, pilotcompute_description, pj_type=None, context=None):
        """ Add a PilotCompute to the PilotComputeService

            Keyword arguments:
            rm -- Contact string for the resource manager
            pilotcompute_description -- PilotCompute Description
            pj_type --  backend type (optional)
            context -- Security context (optional)

            Return value:
            A PilotCompute handle
        """
        pass

    def cancel(self):
        """ Cancel the PilotComputeService.

            This also cancels all the PilotJobs that were under control of this PJS.

            Keyword arguments:
            None

            Return value:
            Result of operation
        """
        pass





#
#  ComputeUnitService
# 
class ComputeUnitService(object):
    """  WorkUnitService.
    
        The WorkUnitService is the application's interface to submit 
        WorkUnits to the Pilot-Manager in the P* Model.

        It can provide the application with a list of WorkUnits that are 
        managed by the Pilot-Manager.

        The WorkUnitService is linked to a PilotComputeService for the actual 
        execution of the WorkUnits.
    """

    def __init__(self, wus_id=None):
        """ Create a Work Service object.
    
            Keyword arguments:
            wus_id -- Reconnect to an existing WUS (optional).
        """
        pass


    def add_pilot_job_service(self, pjs):
        """ Add a PilotComputeService to this WUS.

            Keyword arguments:
            pilotjob_services -- The PilotCompute Service(s) to which this 
                                 Work Unit Service will connect.

            Return:
            Result
        """
        pass


    def remove_pilot_job_service(self, pjs):
        """ Remove a PilotComputeService from this WUS.

            Note that it won't cancel the PilotComputeService, it will just no
            longer be connected to this WUS.

            Keyword arguments:
            pilotjob_services -- The PilotCompute Service(s) to remove from this
                                 Work Unit Service. 

            Return:
            Result
        """
        pass


    def submit_compute_unit(self, wud):
        """ Submit a WU to this Work Unit Service.

            Keyword argument:
            wud -- The WorkUnitDescription from the application

            Return:
            WorkUnit object
        """
        pass


    def cancel(self):
        """ Cancel the WUS.
            
            Cancelling the WUS also cancels all the WUs submitted to it.
    
            Keyword arguments:
            None
    
            Return:
            Result
        """
        pass
    
    def get_state(self):
        pass
    
    def get_id(self):
        pass



#
#  ComuteUnitDescription
# 
class ComputeUnitDescription(dict):
    """  WorkUnitDescription.
    
        The WorkUnitDescription is a job/task/call description based on 
        SAGA Job Description. 
        
        It offers the application to describe a WorkUnit in an abstract 
        way that is dealt with by the Pilot-Manager.
    """

    # Class members
    __slots__ = (
        # Action description
        'executable',           # The "action" to execute
        'arguments',            # Arguments to the "action"
        'cleanup',
        'environment',          # "environment" settings for the "action"
        'interactive', 
        'contact',
        'project',
        'start_time',
        'working_directory',
        # I/O
        'input',
        'error',
        'output',
        'file_transfer',
        # Parallelism
        'number_of_processes',  # Total number of processes to start
        'processes_per_host',   # Nr of processes per host
        'threads_per_process',  # Nr of threads to start per process
        'total_core_count',     # Total number of cores requested
        'spmd_variation',       # Type and startup mechanism
        # Requirements
          'candidate_hosts',
        'cpu_architecture',
        'total_physical_memory',
        'operating_system_type',
        'total_cpu_time',
        'wall_time_limit',
        'queue'
    )

    def __init__(self):
        pass

    def __setattr__(self, attr, value):
        self[attr]=value
        
    
    def __getattr__(self, attr):
        return self[attr]


#
#  ComputeUnit(WU)
# 
class ComputeUnit(object):
    """  WorkUnit.
    
        This is the object that is returned by the WorkUnitService when a 
        new WorkUnit is created based on a WorkUnitDescription.

        The WorkUnit object can be used by the application to keep track 
        of WorkUnits that are active.

        A WorkUnit has state, can be queried and can be cancelled.
    """

    # Class members
    __slots__ = (
        'id',               # Reference to this WU
        'description',      # Description of this WU
        'state_detail',     # Detailed (application specific) state of this WU
        'callback'          # Callback object
    )

    
    def cancel(self):
        """ Cancel the WU. """
        pass

    
    def set_callback(self, member, cb):
        """ Set a callback function for a member.

            Keyword arguments:
            member -- The member to set the callback for (state / state_detail).
            cb -- The callback object to call.
        """
        pass

    
    def unset_callback(self, member):
        """ Unset a callback function from a member

            Keyword arguments:
            member -- The member to unset the callback from.
        """
        pass
    

    def get_state(self):
        pass
    
    
    def wait(self):
        """ Wait until in Done state 
            (or Failed state)
        """
#
#  Callback (Abstract) Class
#
class Callback(object):
    """ Callback class.

        Specifies the structure for callback classes.

        Callbacks can be set for WorkUnits on the state or state_detail members.
    """

    def cb(self, wu, member, value):
        """ This is the method that needs to be implemented by the application
        
            Keyword arguments:
            wu -- The WU that is calling back.
            member -- The member that triggered the callback.
            value -- The new (detailed) state.

            Return:
            Keep -- Keep or remove the callback
        """
        pass



'''
Encapsulates coordination and communication specifics of bigjob
'''
import threading
import datetime
import time
import sys
import os
import pickle
import pdb
import saga
import json
import urlparse
import logging

from bigjob import logger
logger.debug("Load Advert Coordination")

if sys.version_info < (2, 5):
    sys.path.append(os.path.dirname( os.path.abspath( __file__) ) + "/../ext/uuid-1.30/")
    sys.stderr.write("Warning: Using unsupported Python version\n")
    
logging.debug(str(sys.path))
import uuid

APPLICATION_NAME="BigJob/BigJob"
ADVERT_URL_SCHEME = "advert://"
ADVERT_SERVER="advert.cct.lsu.edu"
ADVERT_SERVER_PORT=8080

class bigjob_coordination(object):
    '''
    Encapsulates communication and coordination
    Implementation based on Redis (http://redis.io)
    '''

    def __init__(self, server=ADVERT_SERVER, server_port=ADVERT_SERVER_PORT, 
                       server_connect_url=None, username=None, password=None, 
                       dbtype=None, url_prefix=None):
        '''
        Constructor
        '''
        
        #pdb.set_trace()
        if url_prefix==None:
            url_prefix = ADVERT_URL_SCHEME
             
        if username!=None and username!="":
            url_prefix   = url_prefix+username
            if password!=None:
                url_prefix = url_prefix + ":" + password
            url_prefix = url_prefix + "@"
        if server_connect_url!=None:
            self.address=server_connect_url
        elif server_port != None:
            self.address = url_prefix+"%s:%i"%(server, server_port)
        elif server != None:
            self.address = url_prefix+"%s"%(server)           
        
        self.username=""
        self.password=""
        self.dbtype=""
        surl = saga.url(self.address)
        if server_connect_url==None: # Manager
            if username!=None:
                surl.username=username    
                self.username=username        
            if password != None:
                surl.password = password
                self.password=password
            if dbtype != None:            
                #surl.query = dbtype
                self.dbtype = dbtype
        else: # Agent
            if surl.query!=None:
                self.dbtype=surl.query
                surl.query=""
                
        self.address = str(surl)
        self.pilot_url = self.address
        logger.debug("Server: " + str(server) + " Port " + str(server_port) +
                      " Url prefix: " + str(url_prefix) + 
                      " Address: " + str(self.get_address()) +
                      " server_connect_url: " + str(server_connect_url) )
        logger.debug("Initialized Coordination to: %s (DB: %s)"%(self.address, self.dbtype))
        self.resource_lock = threading.RLock()
        
    
    def get_address(self):
        return self.address + "?" + self.dbtype
    
    
    def get_url(self, id_string):        
        if not id_string.startswith("advert") and not id_string.startswith("sqlasyncadvert"): 
            path = id_string.replace(":", "/")            
            if self.dbtype!=None:
                url_string = self.address + "/" + path + "?" + self.dbtype
            else:
                url_string = self.address + "/" + path 
            return url_string
        
        if self.dbtype!=None:
            id_string = id_string + "?" + self.dbtype
             
        return id_string
        
        
    #####################################################################################
    # Pilot-Job State
    def set_pilot_state(self, pilot_url, new_state, stopped=False):   
        pilot_url = self.get_url(pilot_url)
        logger.debug("create advert entry: " + pilot_url)
        pilot_dir = saga.advert.directory(saga.url(pilot_url), saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        logger.debug("update state of pilot job to: " + str(new_state) + " Stopped: " + str(stopped))
        pilot_dir.set_attribute("state", str(new_state)) 
        pilot_dir.set_attribute("stopped", str(stopped))
        
    def get_pilot_state(self, pilot_url):
        pilot_url = self.get_url(pilot_url)
        pilot_dir = saga.advert.directory(saga.url(pilot_url), saga.advert.Read)
        state = pilot_dir.get_attribute("state")
        stopped = pilot_dir.get_attribute("stopped") 
        if stopped == "false" or stopped == "False":
            return {"state":state, "stopped":False}
        else:
            return {"state":state, "stopped":True}        
        
    #####################################################################################
    # Pilot-Job Description
    def set_pilot_description(self, pilot_url, description):     
        pass
    
    def get_pilot_description(self, pilot_url):
        pass
    
    def get_jobs_of_pilot(self, pilot_url):
        pilot_url = self.get_url(pilot_url + "/jobs")
        """ returns array of job_url that are associated with a pilot """
        pilot_dir = saga.advert.directory(saga.url(pilot_url), saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        jobs = pilot_dir.list()   
        j = [self.__remove_dbtype(pilot_url) + "/" + i.get_string() for i in jobs]
        return j
    
    
    def delete_pilot(self, pilot_url):
        pilot_url = self.get_url(pilot_url)
        pilot_dir = saga.advert.directory(saga.url(pilot_url), saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        pilot_dir.remove(pilot_url, saga.name_space.Recursive)    
    
    
    #####################################################################################
    # Sub-Job State    
    def set_job_state(self, job_url, new_state):   
        self.resource_lock.acquire()     
        job_url = self.get_url(job_url)
        logger.debug("Set state of job: " + str(job_url) + " to: " + str(new_state))
        job_dir = saga.advert.directory(saga.url(job_url), saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        job_dir.set_attribute("state", str(new_state))
        self.resource_lock.release()
        
    def get_job_state(self, job_url):        
        job_url = self.get_url(job_url)        
        job_dir = saga.advert.directory(saga.url(job_url), saga.advert.Read)
        state = job_dir.get_attribute("state")  
        #logger.debug("Get state of job: " + str(job_url) + " state: " + str(state))
        return state      
    
    
    #####################################################################################
    # Sub-Job Description
    def set_job(self, job_url, job_dict):
        job_dir_url = self.get_url(job_url)
        job_description_url = self.get_url(job_url+"/job-description")
        logger.debug("Job URL: %s, Job Description URL: %s"%(job_dir_url, job_description_url))
        #job_dir = saga.advert.directory(saga.url(job_dir_url), 
        #                                saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        # directory is recursively created
        job_desc_entry = saga.advert.entry(saga.url(job_description_url),
                                           saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        logger.debug("initialized advert entry for job: " + job_dir_url)
        job_desc_entry.store_string(json.dumps(job_dict))
        self.set_job_state(job_url, str(saga.job.Unknown))
        
        
    
    def get_job(self, job_url):
        #job_dir = saga.advert.directory(saga.url(job_url), 
        #                                saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        job_url = self.get_url(job_url+"/job-description")
        logger.debug("Get job description from: %s"%(job_url))
        job_desc_entry = saga.advert.entry(saga.url(job_url),
                                           saga.advert.Read)
        job_dict = json.loads(job_desc_entry.retrieve_string())
        return job_dict    
    
    def delete_job(self, job_url):
        job_url = self.get_url(job_url)
        job_dir = saga.advert.directory(saga.url(job_url), 
                                        saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        job_dir.remove(job_url, saga.name_space.Recursive)  
    
    
    #####################################################################################
    # Distributed queue for sub-jobs
    def queue_job(self, pilot_url, job_url):
        self.resource_lock.acquire()
        #pilot_url = self.get_url(pilot_url)
        job_url = self.get_url(job_url)
        """ queue new job to pilot """
        new_job_url = self.get_url(pilot_url + "/new/" + str(uuid.uuid1()))
        logger.debug("Job URL: %s Create new job entry at: %s"%(job_url,new_job_url))
        new_job_dir = saga.advert.directory(saga.url(new_job_url), 
                                            saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        new_job_dir.set_attribute("joburl", job_url)
        self.resource_lock.release()
        
    def dequeue_job(self, pilot_url):
        """ deque to new job  of a certain pilot """
        self.resource_lock.acquire()
        #pilot_url = self.get_url(pilot_url)
        jobs = []        
        new_job_dir_url = self.get_url(pilot_url + "/new/") 
        new_job_dir = saga.advert.directory(saga.url(new_job_dir_url), 
                                            saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        new_jobs = new_job_dir.list()
        logger.debug("Pilot Job base dir: " + new_job_dir_url + " #new jobs: " + str(len(new_jobs))
                      + " jobs: " + str(new_jobs));
        if len(new_jobs)>=1:
            job_entry=new_jobs[0]     
            job_dir_url = self.get_url(pilot_url + "/new/" + "/" + job_entry.get_string())       
            logger.debug("Open job at " + str(job_dir_url))
            job_dir = saga.advert.directory(saga.url(job_dir_url), 
                                       saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
            
            #new_job_dir.open_dir(job_entry)                        
            job_url = job_dir.get_attribute("joburl")
            #remove old job entry
            job_dir.remove(self.__remove_dbtype(job_dir_url), saga.name_space.Recursive)
                
            logger.debug("Dequeued new job: " + str(job_url))
            self.resource_lock.release()
            return self.__remove_dbtype(job_url)
        else:
            self.resource_lock.release()
            time.sleep(1)
            return 
            
    def __remove_dbtype(self, url):
        surl = saga.url(url)
        surl.query=""
        return str(surl)  
            
            



import sys
import getopt
import saga
import time
import pdb
import os
import traceback
import logging
import uuid

import bigjob

from bigjob import logger
from bigjob import bigjob, subjob
from bigjob_dynamic.many_job import *

from pilot.api import PilotCompute, PilotComputeService, State
#from pilot.coordination.advert import AdvertCoordinationAdaptor as CoordinationAdaptor
from pilot.coordination.nocoord import NoCoordinationAdaptor as CoordinationAdaptor


""" This variable defines the coordination system that is used by BigJob
    e.g. 
        advert://localhost (SAGA/Advert SQLITE)
        advert://advert.cct.lsu.edu:8080 (SAGA/Advert POSTGRESQL)
        redis://localhost:6379 (Redis at localhost)
        tcp://localhost (ZMQ)
"""
COORDINATION_URL = "advert://localhost/?dbtype=sqlite3"


class PilotComputeService(PilotComputeService):
    """ PilotJobService based on BigJob.
                    
    """

    PJS_ID_PREFIX="pjs-"   

    # Class members
    __slots__ = (
        'id',           # Reference to this PJS
        'url',
        'state',       # Status of the PJS
        'pilot_computes',    # List of PJs under this PJS
        'coordination_url',
        '__mjs'
    )

    def __init__(self, coordination_url=COORDINATION_URL, pjs_url=None):
        """ Create a PilotJobService object.

            Keyword arguments:
            pjs_id -- Don't create a new, but connect to an existing (optional)
        """
        self.__mjs = None
        self.pilot_computes=[]
        self.coordination_url=coordination_url
        
        if pjs_url==None:      # new pjs          
            self.id = self.PJS_ID_PREFIX+str(uuid.uuid1())
            self.url = "pilotjob://localhost/"+self.id
        else:
            logger.error("Reconnect to PJS currently not supported.")
            

    def create_pilot(self, rm=None, pilot_compute_description=None, pj_type=None, context=None):
        """ Add a PilotJob to the PilotJobService

            Keyword arguments:
            pilot_compute_description -- PilotJob Description
            
            Return value:
            A PilotJob handle
        """
        
        if self.__mjs == None:
            logging.debug("Create Dynamic BigJob Service")            
            self.__mjs = many_job_service([], self.coordination_url)
            
        resource_description = self.__translate_pj_bj_description(pilot_compute_description)
        bigjob = self.__mjs.add_resource(resource_description)
        pj = PilotCompute(self, bigjob, pilot_compute_description)
        self.pilot_computes.append(pj)
        return pj
        
    def __translate_pj_bj_description(self, pilot_compute_description):
        resource_description={}
        if pilot_compute_description.has_key("service_url"):
            resource_description["resource_url"] = pilot_compute_description["service_url"] 
            
        if pilot_compute_description.has_key("queue"):
            resource_description["queue"] = pilot_compute_description["queue"] 
        else:
            resource_description["queue"] = None
            
        if pilot_compute_description.has_key("allocation"):
            resource_description["allocation"] = pilot_compute_description["allocation"] 
        else:
            resource_description["allocation"] = None
        
        for i in pilot_compute_description.keys():
            resource_description[i] = pilot_compute_description[i] 
        
        return resource_description
    
    def list_pilots(self):
        return self.pilot_computes
        
    
    def cancel(self):
        """ Cancel the PilotJobService.

            This also cancels all the PilotJobs that were under control of this PJS.

            Keyword arguments:
            None

            Return value:
            Result of operation
        """
        self.__mjs.cancel()
        
        
    def _submit_cu(self, compute_unit):
        """ Submits work unit to Dynamic Bigjob (ManyJob) 
            Scheduler of Dynamic Bigjob will assign appropriate PJ to WorkUnit        
        """
        subjob = self.__mjs.create_job(compute_unit.subjob_description)
        subjob.run()
        compute_unit.subjob=subjob
        return compute_unit
    
    
    def __repr__(self):
        status_string = "{\n"
        for i in self.pilot_computes:
            status_string = status_string + " PJ: %s, State: %s;"%(i.get_url(), i.get_state())
        status_string = status_string + "\n}"
        return status_string
    
    
    
class PilotCompute(PilotCompute):
    """ Wrapper for BigJob class """
     
    def __init__(self, pilot_compute_service, bigjob, pilot_compute_description):
        logging.debug("Create PilotCompute for BigJob: " + str(bigjob))
        self.pilot_compute_description=pilot_compute_description
        self.__pilot_compute_service=pilot_compute_service
        self.__bigjob = bigjob        
        self.__subjobs = []
        
    def cancel(self):
        self.__bigjob.cancel()
    
    def get_state(self):
        return self.__bigjob.get_state()
    
    def wait(self):
        """ Waits for completion of CUs """
        self.__bigjob.wait()
    
    def get_url(self):
        return self.__bigjob.pilot_url
    
    
    
    def _submit_cu(self, compute_unit):
        """ Submits work unit to Bigjob """
        logging.debug("Submit sub-job to big-job")
        sj = bigjob.bigjob_manager.subjob()
        sj.submit_job(self.__bigjob.pilot_url, compute_unit.subjob_description)
        self.__subjobs.append(sj)
        compute_unit.subjob=sj
        return compute_unit
        
        
        
        



'''
Globus Online based File Transfer
'''
import urlparse
import pdb
import errno
import sys
import os
import stat
import logging
import traceback
import time

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../.."))
#sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
from pilot.api import State
from bigjob import logger

from globusonline.transfer import api_client
from globusonline.transfer.api_client.get_go_cookie import get_go_auth

class GlobusOnlineFileAdaptor(object):
    """ BigData Coordination File Management for Pilot Data """
    
    URL_PREFIX="go://"
    ENDPOINT_DICT={
                  "kraken-gsi.nics.teragrid.org" : "xsede#kraken",
                        
                  }
    
    
    def __init__(self, service_url):        
        self.service_url = service_url
        result = urlparse.urlparse(service_url)
        self.host = result.netloc
        self.query = result.path      
        self.ep = self.__get_ep(self.query)
        self.path = self.__get_path(self.query)
        self.user = result.username  
        self.password = result.password
        
        result = get_go_auth(ca_certs=None, username=self.user, password=self.password)
        saml_cookie = result.cookie
        
        self.api = api_client.TransferAPIClient(username=self.user,
                                                saml_cookie=saml_cookie
                                                )
        status_code, status_message, data = self.api.task_list()
        
        # initialize ssh client
        self.__state=State.New
                        
        
    def initialize_pilotdata(self):
        # check whether directory exists
        self.api.endpoint_mkdir(self.ep, self.path)            
        self.__state=State.Running
        
        
    def get_pilotdata_size(self):
        # check size
        return None
    
    
    def delete_pilotdata(self):
        self.api.endpoint_delete(self.ep, self.path)
        self.__state=State.Done
    
        
    def get_state(self):
        if self.__client.get_transport().is_active()==True:
            return self.__state
        else:
            self.__state=State.Failed
            return self.__state            
            
    def create_du(self, du_id):
        du_dir = os.path.join(self.path, str(du_id))
        logger.debug("mkdir: " + du_dir)
        self.api.endpoint_mkdir(self.ep, du_dir)
        
        
    def put_du(self, du):
        logging.debug("Copy DU using Globus Online")
        for i in du.list_data_unit_items():     
            remote_path = os.path.join(self.path, str(du.id), os.path.basename(i.local_url))
            logging.debug("Put file: %s to %s"%(i.local_url, remote_path))                        
            if i.local_url.startswith("ssh://"):
                # check if remote path is directory
                if self.__is_remote_directory(i.local_url):
                    logging.warning("Path %s is a directory. Ignored."%i.local_url)                
                    continue
                
               
                #self.__third_party_transfer(i.local_url, remote_path)                
            else:
                if stat.S_ISDIR(os.stat(i.local_url).st_mode):
                    logging.warning("Path %s is a directory. Ignored."%i.local_url)                
                    continue         
            result = urlparse.urlparse(i.local_url)
            source_host = result.netloc
            source_path = result.path
            logger.debug(str((source_host, source_path, self.host, remote_path)))
            if source_host == "" or source_host==None:
                cmd = "scp "+ source_path + " " + self.host + ":" + remote_path
            else:
                cmd = "scp "+ source_host+":"+source_path + " " + self.host + ":" + remote_path
            logger.debug("Command: %s"%cmd)
            os.system(cmd)                   
                
    
    def copy_du_to_url(self, du,  local_url, remote_url):
        base_dir = self.__get_path_for_du(du)
        self.__create_remote_directory(remote_url)  
        for filename in self.__sftp.listdir(base_dir):
            file_url = local_url + "/" + filename
            file_remote_url = remote_url + "/" + filename
            logger.debug("Copy " + file_url + " to " + file_remote_url)
            self.__third_party_transfer_host(file_url, file_remote_url)

        

    def copy_du(self, du, pd_new):
        remote_url = pd_new.service_url + "/" + str(du.id)
        local_url =  self.service_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
    
    def get_du(self, du, target_url):
        remote_url = target_url
        local_url =  self.service_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
        
    def remove_du(self, du):
        self.__remove_directory(os.path.join(self.path, du.id))
    
    
    ###########################################################################
    # Pure File Management APIs
        
    def transfer(self, source_url, target_url):
        self.__third_party_transfer_host(source_url, target_url)    
    
    
    def create_remote_directory(self, target_url):
        if not self.__is_remote_directory(target_url):
            result = urlparse.urlparse(target_url)
            target_query = result.path
            target_ep = self.__get_ep(target_query)
            target_path = self.__get_path(target_query)
            result = self.api.endpoint_mkdir(target_ep, target_path)      
            logger.debug("GO EP: %s Directory: %s Creation Result: %s"%(target_ep, target_path, str(result)))      
            #task_id = result[2]["task_id"]
            #logger.debug("Transfer Request Result: %s Task ID: %s"%(str(result), task_id))
            #self.__wait_for_task(task_id)
            return True
        return True
    
    
    def get_path(self, target_url):
        result = urlparse.urlparse(target_url)
        target_query = result.path
        target_path = self.__get_path(target_query)
        return target_path
        
    ###########################################################################
    # Private support methods
    def __get_path_for_du(self, du):
        return os.path.join(self.path, str(du.id))
    

    def __remove_directory(self, path):
        """Remove remote directory that may contain files.        
        """
        if self.__exists(path):
            for filename in self.__sftp.listdir(path):
                filepath = os.path.join(path, filename)
                logging.debug("Delete %s"%filepath)
                if stat.S_ISDIR(self.__sftp.stat(filepath).st_mode):
                    [self.__remove_directory(filepath)]
                else:
                    self.__sftp.remove(filepath)
            self.__sftp.rmdir(path)
            
    
   
            
        
    def __is_remote_directory(self, url):
        try:
            result = urlparse.urlparse(url)
            target_query = result.path
            target_ep = self.__get_ep(target_query)
            target_path = self.__get_path(target_query)
            result = self.api.endpoint_ls(target_ep, target_path)      
            logger.debug("GO EP: %s Directory: %s Creation Result: %s"%(target_ep, target_path, str(result)))
            return True  
        except:
            pass       
        return False
        
    def __create_sftp_client(self):
        ssh_client = paramiko.SSHClient()
        ssh_client.load_system_host_keys()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(self.host)
        sftp_client = ssh_client.open_sftp()
        sftp_client.chdir(self.path)
        return ssh_client, sftp_client
        
        
    def __third_party_transfer_host(self, source_url, target_url):
        """
            Transfers from source URL to machine to target_url
        """
        transfer_start = time.time()
        result = urlparse.urlparse(source_url)
        source_query = result.path
        source_ep = self.__get_ep(source_query)
        source_path = self.__get_path(source_query)
        
        result = urlparse.urlparse(target_url)
        target_query = result.path
        target_ep = self.__get_ep(target_query)
        target_path = self.__get_path(target_query)
        
        transfer_id = self.api.submission_id()[2]["value"]    
        logger.debug("Transfer ID: %s"%transfer_id)    
        transfer = api_client.Transfer(transfer_id, source_ep, target_ep,
                                       deadline=None, sync_level=None, label=None)
        transfer.add_item(source_path=source_path, destination_path=target_path, recursive=False )        
        result = self.api.transfer(transfer)
        task_id = result[2]["task_id"]
        logger.debug("Transfer Request Result: %s Task ID: %s"%(str(result), task_id))
        self.__wait_for_task(task_id)
        logger.debug("Task ID: %s Time: %d sec"%(transfer_id, (time.time()-transfer_start)))
    
    
    def __get_ep(self, query_string):  
        if query_string.startswith("?"):
            query_string = query_string[1:]      
        comp = query_string.split("&")
        for i in comp:
            part = i.split("=")
            if part[0]=="ep":
                return part[1]                
    
    def __get_path(self, query_string):
        if query_string.startswith("?"):
            query_string = query_string[1:]      
        comp = query_string.split("&")
        for i in comp:
            part = i.split("=")
            if part[0]=="path":
                return part[1]      
    
    def __wait_for_task(self, task_id, timeout=None):
        status = "ACTIVE"
        while (timeout==None or timeout > 0) and status == "ACTIVE":
            code, reason, data = self.api.task(task_id, fields="status")
            status = data["status"]
            time.sleep(1)
            if timeout!=None:
                timeout -= 1

        if status != "ACTIVE":
            print "Task %s complete!" % task_id
            return True
        else:
            print "Task still not complete after %d seconds" % timeout
            return False
    
    
    def __exists(self, path):
        """Return True if the remote path exists
        """
        try:
            self.__sftp.stat(path)
        except IOError, e:
            if e.errno == errno.ENOENT:
                return False
            raise
        else:
            return True
   
   
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    
    
if __name__ == "__main__":
    go = GlobusOnlineFileAdaptor("http://drelu:bigjob@cli.globusonline.org?ep=drelu#egi&path=/ho")
    go.transfer("go://cli.globusonline.org?ep=drelu#MacBook&path=/~/cert.tar.gz", "go://cli.globusonline.org?ep=xsede#kraken&path=/~/")
    


import os
import logging

version = "latest"

#READ config
SAGA_BLISS=False
try:
    import ConfigParser
    _CONFIG_FILE="bigjob.conf"
    _conf_file = os.path.dirname(os.path.abspath( __file__ )) + "/../" + _CONFIG_FILE
    _config = ConfigParser.ConfigParser()
    _config.read(_conf_file)
    default_dict = _config.defaults()
    
    level = default_dict["logging.level"]
    print("Logging level: %s"%level) 
    if level.startswith("logging."):
        logging_level = eval(level)        
        print("Set logging level: %s/%s"%(level,logging_level))
        logging.basicConfig(level=logging_level, datefmt='%m/%d/%Y %I:%M:%S %p',
                   format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    else:
        logging.basicConfig(level=logging.DEBUG, datefmt='%m/%d/%Y %I:%M:%S %p',
                   format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')    
        
    logger = logging.getLogger(name='bigjob')
            
    saga = default_dict["saga"]
    if saga.lower() == "bliss":
        SAGA_BLISS=True    
        
except:
    print("bjgjob.conf could not be read") 
    

try:
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", 'VERSION')
    version = open(fn).read().strip()
    logger.info("Loading BigJob version: " + version)
except IOError:
    pass



# define external-facing API
from bigjob.bigjob_manager import bigjob as myBigjob
from bigjob.bigjob_manager import subjob as mySubjob
from bigjob.bigjob_manager import description as myDescription

class subjob(mySubjob):
    pass


class bigjob(myBigjob):
    pass


class description(myDescription):
    pass



""" PilotData 
""" 
import sys
import os
import logging
logging.basicConfig(level=logging.DEBUG)

import uuid
import random
import threading
import time
import pdb
import Queue

import saga
sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from bigjob import logger

from pilot.api import PilotData, DataUnit, PilotDataService
from pilot.api import State


from pilot.filemanagement.ssh_adaptor import SSHFileAdaptor 
try:
    from pilot.filemanagement.webhdfs_adaptor import WebHDFSFileAdaptor
except:
    logger.warn("WebHDFS package not found.") 
try:
    from pilot.filemanagement.globusonline_adaptor import GlobusOnlineFileAdaptor
except:
    logger.warn("Globus Online package not found.") 



#from pilot.coordination.advert import AdvertCoordinationAdaptor as CoordinationAdaptor
from pilot.coordination.nocoord import NoCoordinationAdaptor as CoordinationAdaptor


# generate global application id for this instance
#application_id = str(uuid.uuid1())
application_id = "bigdata"

class PilotData(PilotData):
    """ PilotData. 
    
        Reserves a space of physical storage on the resource specified in the pilot_data_description
    """   
    
    PD_ID_PREFIX="pd-"   

        
    def __init__(self, pilot_data_service=None, pilot_data_description=None, pd_url=None):    
        """ 
            Initialize PilotData at given service url:
            
            ssh://<hostname>
            gsissh://<hostname>
            
            Currently only ssh schemes are supported. In the future all 
            SAGA URL schemes/adaptors should be supported.        
        """ 
        self.id = None
        self.url = None
        self.pilot_data_description = None
        self.service_url=None
        self.size = None
        self.data_unit_description = None
        self.data_units={}
        
        if pd_url==None and pilot_data_service!=None:      # new pd          
            self.id = self.PD_ID_PREFIX+str(uuid.uuid1())
            self.pilot_data_description = pilot_data_description
            self.url = CoordinationAdaptor.add_pd(CoordinationAdaptor.get_base_url(application_id)+"/"+pilot_data_service.id, self)
        elif pd_url != None:
            logger.warn("Reconnect to PilotData: %s"%pd_url)
            dictionary = CoordinationAdaptor.get_pd(pd_url)
            pd_dict = dictionary["pilot_data"]
            for i in pd_dict:
                self.__setattr__(i, pd_dict[i])
                        
        self.initialize_pilot_data()
        
            
    def initialize_pilot_data(self):
        if self.pilot_data_description!=None:
            self.service_url=self.pilot_data_description["service_url"]
            self.size = self.pilot_data_description["size"]
            
            # initialize file adaptor
            if self.service_url.startswith("ssh:"):
                logger.debug("Use SSH backend")
                self.__filemanager = SSHFileAdaptor(self.service_url)
            elif self.service_url.startswith("http:"):
                logger.debug("Use WebHDFS backend")
                self.__filemanager = WebHDFSFileAdaptor(self.service_url)
                
            self.__filemanager.initialize_pilotdata()
            self.__filemanager.get_pilotdata_size()
            

    def __get_pd_id(self, pd_url):
        start = pd_url.index(self.PD_ID_PREFIX)
        end =pd_url.index("/", start)
        return pd_url[start:end]
    

    def cancel(self):        
        """ Cancel PilotData 

            Keyword arguments:
            None
        """
        #self.__filemanager.delete_pilotdata()
        pass
        
        
    def url_for_du(self, du):
        if self.data_units.has_key(du.id):
            return self.service_url + "/" + str(du.id)
        return None
    
    
    def create_du(self, du):
        self.__filemanager.create_du(du.id)
        
        
    def put_du(self, du):
        logging.debug("Put PD: %s to PS: %s"%(du.id,self.service_url))
        self.__filemanager.create_du(du.id)
        self.__filemanager.put_du(du)
        self.data_units[du.id] = du
        CoordinationAdaptor.update_pd(self)
        
        
    def remove_du(self, du):
        """ Remove pilot data from pilot data """
        if self.data_units.has_key(du.id):
            self.__filemanager.remove_du(du)
            del self.data_units[du.id]
        CoordinationAdaptor.update_pd(self)
        
    
    def copy_du(self, du, pd_new):
        pd_new.create_du(du)
        self.__filemanager.copy_du(du, pd_new)
        
        # update meta data at pd_new
        pd_new.data_units[du.id] = du
        CoordinationAdaptor.update_pd(pd_new)
        
    
    def list_data_units(self):
        return self.data_units.values()           
        #return self.data_units.values()
    
    
    def get_state(self):
        return self.__filemanager.get_state()
    
    
    def wait(self):
        while 1:
            finish_counter=0
            result_map = {}
            dus = self.data_units.values()
            for du in dus: 
                du.wait()
                state = du.get_state()           
                #state = job_detail["state"]                
                if result_map.has_key(state)==False:
                    result_map[state]=1
                else:
                    result_map[state] = result_map[state]+1
                if self.__has_finished(state)==True:
                    finish_counter = finish_counter + 1                   
            logger.debug("PD ID: " + str(self.id) + " Total DUs: %s States: %s"%(len(dus), str(result_map)))
            if finish_counter == len(dus):
                break
            time.sleep(2)

    
    def export_du(self, du, target_url):
        self.__filemanager.get_du(du, target_url)
    
    
    def to_dict(self):
        pd_dict = {}
        pd_dict["id"]=self.id
        pd_dict["url"]=self.url
        pd_dict["pilot_data_description"]=self.pilot_data_description
        logger.debug("PS Dictionary: " + str(pd_dict))
        return pd_dict
    
    
    def __repr__(self):
        return self.service_url
    
    
    def __has_finished(self, state):
        state = state.lower()
        if state=="running" or state=="failed" or state=="canceled":
            return True
        else:
            return False
    
    @classmethod
    def create_pilot_data_from_dict(cls, pd_dict):
        pd = PilotData()
        for i in pd_dict.keys():
            pd.__setattr__(i, pd_dict[i])
        pd.initialize_pilot_data()
        logger.debug("created pd " + str(pd))
        return pd
    

class PilotDataService(PilotDataService):
    """ PilotDataService (PSS)."""
    
    PDS_ID_PREFIX="pds-"

    # Class members
    __slots__ = (
        'id',             # Reference to this PJS
        'url',            # URL for referencing PilotDataService
        'state',          # Status of the PJS
        'data_unit'    # List of PJs under this PJS
        'affinity_list'   # List of PS on that are affine to each other
    )

    def __init__(self, pds_url=None):
        """ Create a PilotDataService

            Keyword arguments:
            pds_id -- restore from pds_id
        """        
        self.pilot_data={}
        
        if pds_url == None:
            self.id = self.PDS_ID_PREFIX + str(uuid.uuid1())
            application_url = CoordinationAdaptor.get_base_url(application_id)
            self.url = CoordinationAdaptor.add_pds(application_url, self)
        else:
            self.id = self.__get_pds_id(pds_url)
    
    
    def __get_pds_id(self, pds_url):
        start = pds_url.index(self.PDS_ID_PREFIX)
        end =pds_url.index("/", start)
        return pds_url[start:end]
    
    def __restore_pd(self, pds_url):
        pd_list=CoordinationAdaptor.list_pd(pds_url) 
        for i in pd_list:
            pass

    def create_pilot(self, pilot_data_description):
        """ Create a PilotData 

            Keyword arguments:
            pilot_data_description -- PilotData Description    
            {
                'service_url': "ssh://<hostname>/base-url/"                
                'size': "1000"
            }
            Return value:
            A PilotData handle
        """
        pd = PilotData(pilot_data_service=self, 
                        pilot_data_description=pilot_data_description)
        self.pilot_data[pd.id]=pd
        
        # store pilot data in central data space
        CoordinationAdaptor.add_pd(self.url, pd)        
        return pd
    
    
    def get_pilot(self, pd_id):
        if self.pilot_data.has_key(pd_id):
            return self.pilot_data[pd_id]
        return None


    def list_pilots(self):
        """ List all PSs of PSS """
        return self.pilot_data.values()
    

    def cancel(self):
        """ Cancel the PilotDataService.
            
            Keyword arguments:
            None

            Return value:
            Result of operation
        """
        for i in self.pilot_data.values():
            i.cancel()
 
 
    def wait(self):
        for i in self.pilot_data.values():
            i.wait()
 
    
    def to_dict(self):
        pds_dict = self.__dict__
        pds_dict["id"]=self.id
        return pds_dict
 
 
    def __del__(self):
        self.cancel()         
            
    

class DataUnit(DataUnit):
    """ Holds a set of file
    
        State model:
            New: PD object created
            Pending: PD object is currently updated  
            Running: At least 1 replica of PD is persistent in a pilot data            
    """
    
    DU_ID_PREFIX="du-"  

    def __init__(self, pilot_data_service=None, data_unit_description=None, du_url=None):
        """
            1.) create a new Pilot Data: pilot_data_service and data_unit_description required
            2.) reconnect to an existing Pilot Data: du_url required 
            
        """
        if du_url==None:
            self.id = self.DU_ID_PREFIX + str(uuid.uuid1())
            self.data_unit_description = data_unit_description        
            self.pilot_data=[]
            self.url = CoordinationAdaptor.add_du(pilot_data_service.url, self)
            self.state = State.New
            self.data_unit_items = DataUnitItem.create_data_unit_list(self, self.data_unit_description["file_urls"]) 
            CoordinationAdaptor.update_du(self)
        else:
            self.id = self.__get_du_id(du_url)
            self.url = du_url   
            logger.debug("Restore du: %s"%self.id)         
            self.__restore_state()
            
        self.transfer_threads=[]
    
    
    def __restore_state(self):
        du_dict = CoordinationAdaptor.get_du(self.url)
        self.data_unit_description = du_dict["data_unit_description"]
        self.state = du_dict["state"]
        data_unit_dict_list = du_dict["data_units"]
        self.data_unit_items = [DataUnitItem.create_data_unit_from_dict(i) for i in data_unit_dict_list]
        
        # restore pilot data
        self.pilot_data = [] 
        for i in du_dict["pilot_data"]:
            logger.debug("PD: "+str(i)) 
            pd = PilotData(pd_url=str(i))
            self.pilot_data.append(pd) 
            
            
    def cancel(self):
        """ Cancel the DU. """
        self.state = State.Done    
        CoordinationAdaptor.update_du(self)
            
    def add_data_unit(self, data_unit):
        self.data_unit_items.append(data_unit)    
        CoordinationAdaptor.update_du(self)
        # TODO Update Pilot Data
        
        
    def remove_data_unit(self, data_unit):
        self.data_unit_items.remove(data_unit)
        CoordinationAdaptor.update_du(self)
        # TODO Update Pilot Data
        
        
    def list_data_unit_items(self):        
        return self.data_unit_items
        
    
    def get_state(self):        
        return self.state  
    
    
    def wait(self):
        """ Wait until in running state 
            (or failed state)
        """
        logger.debug("DU: %s wait()"%(str(self.id)))
        # Wait for all transfers to finish
        for i in self.transfer_threads:
            i.join()
        
        # Wait for state to change
        while self.state!=State.Running and self.state!=State.Failed:
            logger.debug("State: %s"%self.state)
            time.sleep(2)
    
    
    def add_pilot_data(self, pilot_data):
        """ add DU to a certain pilot data 
            data will be moved into this data
        """
        transfer_thread=threading.Thread(target=self.__add_pilot_data, args=[pilot_data])
        transfer_thread.start()        
        self.transfer_threads.append(transfer_thread)
        
    
    def get_pilot_data(self):
        """ get a list of pilot data that have a copy of this PD """
        return self.pilot_data
    
    
    def export(self, target_url):
        """ simple implementation of export: 
                copies file from first pilot data to local machine
        """
        if len(self.pilot_data) > 0:
            self.pilot_data[0].export_du(self, target_url)
        else:
            logger.error("No Pilot Data for PD found")
    
    
    def to_dict(self):
        du_dict = self.__dict__
        du_dict["id"]=self.id
        return du_dict        
    
    
    ###########################################################################
    # BigData Internal Methods
    def update_state(self, state):
        self.state=state
        CoordinationAdaptor.update_du(self)

    
    def __add_pilot_data(self, pilot_data):
        logger.debug("add du to pilot data")
        if len(self.pilot_data) > 0: # copy files from other pilot data
            self.pilot_data[0].copy_du(self, pilot_data)
        else: # copy files from original location
            pilot_data.put_du(self)
        self.pilot_data.append(pilot_data)
        CoordinationAdaptor.update_du(self)  
        
        
    def __get_du_id(self, du_url):
        try:
            start = du_url.index(self.DU_ID_PREFIX)
            end = du_url.find("/", start)
            if end==-1:
                end = du_url.find("?", start)
            if end==-1:
                end = len(du_url)-1
            return du_url[start:end]
        except:
            logger.error("No valid PD URL")
        return None
    
    
    def __repr__(self):        
        return "PD: " + str(self.url) 
        + " \nData Units: " + str(self.data_unit_items)
        + " \nPilot Stores: " + str(self.pilot_data)
    
    

class DataUnitItem(object):
    """ DataUnitItem """
    DUI_ID_PREFIX="dui-"  
   
    def __init__(self, pd=None, local_url=None):        
        if local_url!=None:
            self.id = self.DUI_ID_PREFIX + str(uuid.uuid1())
            self.local_url = local_url        
            if pd != None:
                self.url = pd.url + os.path.basename(local_url)
        
        
    @classmethod    
    def __exists_file(cls, url):   
        """ return True if file at url exists. Otherwise False """
        file_url = saga.url(url)
        if file_url.host == "":
            if os.path.exists(str(file_url)):
                return True
            else:
                return False            
        elif file_url.host=="localhost":
            if os.path.exists(file_url.path):
                return True
            else:
                return False
        else:            
            return True
        
    def to_dict(self):
        du_dict = self.__dict__
        du_dict["id"]=self.id
        return du_dict
    
    
    def __repr__(self):
        return str(self.__dict__) 
        
         
    @classmethod
    def create_data_unit_list(cls, pd=None, urls=None):
        """ Creates a list of DUs from URL list
        """    
        du_list = []    
        for i in urls:            
            if cls.__exists_file(i):
                du = DataUnitItem(pd, i)
                du_list.append(du)
    
        return du_list
    
    @classmethod
    def create_data_unit_from_urls(cls, pd=None, urls=None):
        """ Creates a list of DUs from URL list
        """    
        du_list = []    
        for i in urls:            
            if cls.__exists_file(i):
                du = DataUnitItem(pd, i)
                du_list.append(du)
    
        return du_list
    
    
    @classmethod
    def create_data_unit_from_dict(cls, du_dict):
        du = DataUnitItem()
        logger.debug("Restore DU: " + str(du_dict))
        for i in du_dict.keys():
            logger.debug("Set attribute: %s", i)
            du.__setattr__(i, du_dict[i])
        return du
    
    



""" Example application demonstrating job submission via bigjob 

    DON'T EDIT THIS FILE (UNLESS THERE IS A BUG)
    
    THIS FILE SHOULD NOT BE COMMITTED TO SVN WITH USER-SPECIFIC PATHS!
"""

import os
import time
import pdb
# BigJob implementation can be swapped here by importing another implementation,
# e.g. condor, cloud, azure
import sys
sys.path.insert(0, os.getcwd() + "/../")

# configuration
""" This variable defines the coordination system that is used by BigJob
    e.g. 
        advert://localhost (SAGA/Advert SQLITE)
        advert://advert.cct.lsu.edu:8080 (SAGA/Advert POSTGRESQL)
        advert://advert.cct.lsu.edu:5432 (SAGA/Advert POSTGRESQL)
        redis://localhost:6379 (Redis at localhost)
        tcp://localhost (ZMQ)
        tcp://* (ZMQ - listening to all interfaces)
"""

### EDIT COORDINATION_URL to point to advert server.  
COORDINATION_URL = "advert://localhost/?dbtype=sqlite3"
#COORDINATION_URL = "redis://localhost"
#COORDINATION_URL = "advert://advert.cct.lsu.edu:8080/"

from bigjob import bigjob, subjob, description


### EDIT based on the number of jobs you want to submit
NUMBER_JOBS=8

def has_finished(state):
        state = state.lower()
        if state=="done" or state=="failed" or state=="canceled":
            return True
        else:
            return False


""" Test Job Submission via Advert """
if __name__ == "__main__":

    starttime=time.time()

    ##########################################################################################
    # Edit parameters for BigJob
    queue=None # if None default queue is used
    project=None # if None default allocation is used 
    walltime=100
    processes_per_node=4
    number_of_processes =2
    workingdirectory= os.path.join(os.getcwd(), "agent") # working directory for agent
    userproxy = None # userproxy (not supported yet due to context issue w/ SAGA)


    """ 
    URL of the SAGA Job Service that is used to dispatch the pilot job.
    The following URLs are accepted:
    
    lrms_url = "gram://oliver1.loni.org/jobmanager-pbs" # globus resource url used when globus is used. (LONI)
    lrms_url = "pbspro://louie1.loni.org" # pbspro resource url used when pbspro scheduling system is used.(Futuregrid or LSU Machines)
    lrms_url = "ssh://louie1.loni.org" # ssh resource url which launches jobs on target machine. Jobs not submitted to scheduling system.
    lrms_url = "pbs-ssh://louie1.loni.org" # Submit jobs to scheduling system of remote machine.
    lrms_url = "xt5torque://localhost" # torque resource url.
    
    Please ensure that the respective SAGA adaptor is installed and working
    """
    lrms_url = "fork://localhost" 
    
    ##########################################################################################

    print "Start Pilot Job/BigJob at: " + lrms_url
    bj = bigjob(COORDINATION_URL)
    bj.start_pilot_job( lrms_url,
                        None,
                        number_of_processes,
                        queue,
                        project,
                        workingdirectory,
                        userproxy,
                        walltime,
                        processes_per_node)

    print "Pilot Job/BigJob URL: " + bj.pilot_url + " State: " + str(bj.get_state())

    ##########################################################################################
    # Submit SubJob through BigJob
    jobs = []
    job_start_times = {}
    job_states = {}
    for i in range(0, NUMBER_JOBS):
        jd = description()
        jd.executable = "/bin/date"
        jd.number_of_processes = "1"
        jd.spmd_variation = "single"
        jd.arguments = [""]
        jd.output = "sj-stdout-"+str(i)+".txt"
        jd.error = "sj-stderr-"+str(i)+".txt"

        sj = subjob()
        sj.submit_job(bj.pilot_url, jd)
        jobs.append(sj)
        job_start_times[sj]=time.time()
        job_states[sj] = sj.get_state()

    print "Terminating application. You can reconnect to BJ via the following URL: %s"%bj.pilot_url




'''
SSH-based coordination scheme between manager and agent
'''
import paramiko
import urlparse
import pdb
import errno
import sys
import os
import stat
import logging
import traceback

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot.api import State
from bigjob import logger


class SSHFileAdaptor(object):
    """ BigData Coordination File Management for Pilot Store """
    
    def __init__(self, service_url):        
        self.service_url = service_url
        result = urlparse.urlparse(service_url)
        self.host = result.netloc
        self.path = result.path        
        
        # initialize ssh client
        self.__client = paramiko.SSHClient()
        self.__client.load_system_host_keys()
        self.__client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.__client.connect(self.host)
        self.__sftp = self.__client.open_sftp()
        self.__state=State.New
                        
        
    def initialize_pilotdata(self):
        # check whether directory exists
        try:
            self.__sftp.chdir(self.path)            
        except IOError:
            # directory does not exist
            self.__sftp.mkdir(self.path)        
        self.__state=State.Running
        
        
    def get_pilotdata_size(self):
        # check size
        size = self.__sftp.stat(self.path).st_size
        return size
    
    
    def delete_pilotdata(self):
        self.__remove_directory(self.path)
        self.__state=State.Done
    
        
    def get_state(self):
        if self.__client.get_transport().is_active()==True:
            return self.__state
        else:
            self.__state=State.Failed
            return self.__state            
            
    def create_du(self, du_id):
        du_dir = os.path.join(self.path, str(du_id))
        logger.debug("mkdir: " + du_dir)
        try:
            self.__sftp.mkdir(du_dir)
        except:
            pass # dir already exists
        
        
    def put_du(self, du):
        self.put_du_scp(du)
                
                
    def put_du_paramiko(self, du):
        logging.debug("Copy DU using Paramiko")
        for i in du.list_data_unit_items():     
            remote_path = os.path.join(self.path, str(du.id), os.path.basename(i.local_url))
            logging.debug("Put file: %s to %s"%(i.local_url, remote_path))
                        
            if i.local_url.startswith("ssh://"):
                # check if remote path is directory
                if self.__is_remote_directory(i.local_url):
                    logging.warning("Path %s is a directory. Ignored."%i.local_url)                
                    continue      
                self.__third_party_transfer(i.local_url, remote_path)                
            else:
                if stat.S_ISDIR(os.stat(i.local_url).st_mode):
                    logging.warning("Path %s is a directory. Ignored."%i.local_url)                
                    continue            
                #self.__sftp.put(i.local_url, remote_path, self.put_progress, True)
                ssh_client, sftp_client = self.__create_sftp_client()
                sftp_client.put(i.local_url, remote_path)
                sftp_client.close()
                ssh_client.close()           


    def put_du_scp(self, du):
        logging.debug("Copy DU using SCP")
        for i in du.list_data_unit_items():     
            remote_path = os.path.join(self.path, str(du.id), os.path.basename(i.local_url))
            logging.debug("Put file: %s to %s"%(i.local_url, remote_path))                        
            if i.local_url.startswith("ssh://"):
                # check if remote path is directory
                if self.__is_remote_directory(i.local_url):
                    logging.warning("Path %s is a directory. Ignored."%i.local_url)                
                    continue
                
               
                #self.__third_party_transfer(i.local_url, remote_path)                
            else:
                if stat.S_ISDIR(os.stat(i.local_url).st_mode):
                    logging.warning("Path %s is a directory. Ignored."%i.local_url)                
                    continue         
            result = urlparse.urlparse(i.local_url)
            source_host = result.netloc
            source_path = result.path
            logger.debug(str((source_host, source_path, self.host, remote_path)))
            if source_host == "" or source_host==None:
                cmd = "scp "+ source_path + " " + self.host + ":" + remote_path
            else:
                cmd = "scp "+ source_host+":"+source_path + " " + self.host + ":" + remote_path
            logger.debug("Command: %s"%cmd)
            os.system(cmd)                   
                
    
    def copy_du_to_url(self, du,  local_url, remote_url):
        base_dir = self.__get_path_for_du(du)
        self.create_remote_directory(remote_url)  
        for filename in self.__sftp.listdir(base_dir):
            file_url = local_url + "/" + filename
            file_remote_url = remote_url + "/" + filename
            logger.debug("Copy " + file_url + " to " + file_remote_url)
            self.__third_party_transfer_host(file_url, file_remote_url)

        

    def copy_du(self, du, pd_new):
        remote_url = pd_new.service_url + "/" + str(du.id)
        local_url =  self.service_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
    
    def get_du(self, du, target_url):
        remote_url = target_url
        local_url =  self.service_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
        
    def remove_du(self, du):
        self.__remove_directory(os.path.join(self.path, du.id))
    
        
    def put_progress(self, transfered_bytes, total_bytes):
        logging.debug("Bytes transfered %d/%d"%(transfered_bytes, total_bytes))
    
        
    
    ####################################################################################
    # pure file management methods
    # used by BJ file staging
    def transfer(self, source_url, target_url):
        self.__third_party_transfer_host(source_url, target_url)    
    
    def create_remote_directory(self, target_url):
        result = urlparse.urlparse(target_url)
        target_host = result.netloc
        target_path = result.path
        try:
            if not self.__is_remote_directory(target_url):
                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                client.connect(target_host)
                sftp = client.open_sftp()  
                sftp.mkdir(target_path)
                sftp.close()
                client.close()
                return True
        except:
            logger.error("Error creating directory: " + str(target_path) 
                         + " at: " + str(target_host))
            self.__print_traceback()
        return False
        
    def get_path(self, target_url):
        result = urlparse.urlparse(target_url)
        return result.path
        
    ###########################################################################
    # Private support methods
    def __get_path_for_du(self, du):
        return os.path.join(self.path, str(du.id))
    
    def __remove_directory(self, path):
        """Remove remote directory that may contain files.        
        """
        if self.__exists(path):
            for filename in self.__sftp.listdir(path):
                filepath = os.path.join(path, filename)
                logging.debug("Delete %s"%filepath)
                if stat.S_ISDIR(self.__sftp.stat(filepath).st_mode):
                    [self.__remove_directory(filepath)]
                else:
                    self.__sftp.remove(filepath)
            self.__sftp.rmdir(path)
            
  
            
        
    def __is_remote_directory(self, url):
        result = urlparse.urlparse(url)
        host = result.netloc
        path = result.path
        #if path.endswith("/"):
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(host)
        sftp = client.open_sftp()
        try:
            if stat.S_ISDIR(sftp.stat(path).st_mode):
                return True
            else:
                return False
        except:
            logger.error("Directory not found: %s"%path)
        sftp.close()
        client.close()
        return False
        
    def __create_sftp_client(self):
        ssh_client = paramiko.SSHClient()
        ssh_client.load_system_host_keys()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(self.host)
        sftp_client = ssh_client.open_sftp()
        sftp_client.chdir(self.path)
        return ssh_client, sftp_client
        
    def __third_party_transfer_host(self, source_url, target_url):
        """
            Transfers from source URL to machine of PS (target path)
        """
        result = urlparse.urlparse(source_url)
        source_host = result.netloc
        source_path = result.path
        
        result = urlparse.urlparse(target_url)
        target_host = result.netloc
        target_path = result.path
          
        python_script= """import sys
import os
import urllib
import sys
import time
import paramiko

client = paramiko.SSHClient()
client.load_system_host_keys()
client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
client.connect("%s")
sftp = client.open_sftp()
sftp.put("%s", "%s")
"""%(target_host, source_path, target_path)

        logging.debug("Execute: \n%s"%python_script)
        source_client = paramiko.SSHClient()
        source_client.load_system_host_keys()
        source_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        source_client.connect(source_host)
        stdin, stdout, stderr = source_client.exec_command("python -c \'%s\'"%python_script)
        stdin.close()
        logging.debug("************************************************")
        logging.debug("Stdout: %s\nStderr:%s", stdout.read(), stderr.read())
        logging.debug("************************************************")
        
    
    def __third_party_transfer(self, source_url, target_path):
        """
            Transfers from source URL to machine of PS (target path)
        """
        result = urlparse.urlparse(source_url)
        source_host = result.netloc
        source_path = result.path
        
        python_script= """import sys
import os
import urllib
import sys
import time
import paramiko

client = paramiko.SSHClient()
client.load_system_host_keys()
client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
client.connect("%s")
sftp = client.open_sftp()
sftp.get("%s", "%s")

"""%(source_host, source_path, target_path)

        logging.debug("Execute: \n%s"%python_script)
        source_client = paramiko.SSHClient()
        source_client.load_system_host_keys()
        source_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        source_client.connect(self.host)
        stdin, stdout, stderr = source_client.exec_command("python -c \'%s\'"%python_script)
        stdin.close()
        logging.debug("************************************************")
        logging.debug("Stdout: %s\nStderr:%s", stdout.read(), stderr.read())
        logging.debug("************************************************")
    
    def __exists(self, path):
        """Return True if the remote path exists
        """
        try:
            self.__sftp.stat(path)
        except IOError, e:
            if e.errno == errno.ENOENT:
                return False
            raise
        else:
            return True
   
   
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    


#!/usr/bin/env python

"""Module bigjob_manager.

This Module is used to launch jobs via the advert service. 

It assumes that an bigjob_agent.py is available on the remote machine.
bigjob_agent.py will poll the advert service for new jobs and run these jobs on the respective
machine .

Background: This approach avoids queueing delays since the igjob_agent_launcher.py must be just started via saga.job or saga.cpr
once. All shortrunning task will be started using the protocol implemented by subjob() and bigjob_agent.py

Installation:
Set environment variable BIGJOB_HOME to installation directory
"""

import sys
from bigjob import logger
import time
import os
import traceback
import logging
import textwrap
import urlparse
import subprocess

from bigjob import SAGA_BLISS 
from bigjob.state import Running, New, Failed, Done, Unknown

if SAGA_BLISS == False:
    try:
        import saga
        logger.debug("Using SAGA C++/Python.")
        is_bliss=False
    except:
        logger.warn("SAGA C++ and Python bindings not found. Using Bliss.")
        try:
            import bliss.sagacompat as saga
            is_bliss=True
        except:
            logger.warn("SAGA Bliss not found")
else:
    logger.debug("Using SAGA Bliss.")
    try:
        import bliss.sagacompat as saga
        is_bliss=True 
    except:
        logger.warn("SAGA Bliss not found")


# import other BigJob packages
# import API
import api.base
sys.path.append(os.path.dirname(__file__))

from pbsssh import pbsssh
from sgessh import sgessh



#
#try:
#    import paramiko
#except:
#    logger.warn("Paramiko not found. Without Paramiko file staging is not supported!")


if sys.version_info < (2, 5):
    sys.path.append(os.path.dirname( __file__ ) + "/ext/uuid-1.30/")
    sys.stderr.write("Warning: Using unsupported Python version\n")
if sys.version_info < (2, 4):
    sys.path.append(os.path.dirname( __file__ ) + "/ext/subprocess-2.6.4/")
    sys.stderr.write("Warning: Using unsupported Python version\n")
if sys.version_info < (2, 3):
    sys.stderr.write("Error: Python versions <2.3 not supported\n")
    sys.exit(-1)

import uuid

def get_uuid():
    wd_uuid=""
    wd_uuid += str(uuid.uuid1())
    return wd_uuid


""" Config parameters (will move to config file in future) """
CLEANUP=True

#for legacy purposes and support for old BJ API
pilot_url_dict={} # stores a mapping of pilot_url to bigjob



class BigJobError(Exception):
    def __init__(self, value):
        self.value = value
    
    def __str__(self):
        return repr(self.value)
    

class bigjob(api.base.bigjob):
    
    __APPLICATION_NAME="bigjob" 
    
    def __init__(self, coordination_url="advert://localhost/", pilot_url=None):    
        """ Initializes BigJob's coordination system
            e.g.:
            advert://localhost (SAGA/Advert SQLITE)
            advert://advert.cct.lsu.edu:8080 (SAGA/Advert POSTGRESQL)
            redis://localhost:6379 (Redis at localhost)
            tcp://localhost (ZMQ)
        """  
        self.coordination_url = coordination_url
        self.coordination = self.__init_coordination(coordination_url)
        self.launch_method=""
        self.__filemanager=None
        
        # restore existing BJ or initialize new BJ
        if pilot_url!=None:
            logger.debug("Reconnect to BJ: %s"%pilot_url)
            self.pilot_url=pilot_url
            self.uuid = self.__get_bj_id(pilot_url)
            self.app_url = self.__APPLICATION_NAME +":" + str(self.uuid)
            self.job = None
            self.working_directory = None
            self.state=self.get_state_detail()
            pilot_url_dict[self.pilot_url]=self
        else:
            self.uuid = "bj-" + str(get_uuid())        
            logger.debug("init BigJob w/: " + coordination_url)
            self.app_url =self. __APPLICATION_NAME +":" + str(self.uuid) 
            self.state=Unknown
            self.pilot_url=""
            self.job = None
            self.working_directory = None
            logger.debug("initialized BigJob: " + self.app_url)
        
       
        
       
    def __get_bj_id(self, pilot_url):
        start = pilot_url.index("bj-")
        end =pilot_url.index(":", start)
        return pilot_url[start:end]
    
     
    def __init_coordination(self, coordination_url):        
        if(coordination_url.startswith("advert://") or coordination_url.startswith("sqlasyncadvert://")):
            try:
                from coordination.bigjob_coordination_advert import bigjob_coordination
                logger.debug("Utilizing ADVERT Backend")
            except:
                logger.error("Advert Backend could not be loaded")
        elif (coordination_url.startswith("redis://")):
            try:
                from coordination.bigjob_coordination_redis import bigjob_coordination      
                logger.debug("Utilizing Redis Backend")
            except:
                logger.error("Error loading pyredis.")
        elif (coordination_url.startswith("tcp://")):
            try:
                from coordination.bigjob_coordination_zmq import bigjob_coordination
                logger.debug("Utilizing ZMQ Backend")
            except:
                logger.error("ZMQ Backend not found. Please install ZeroMQ (http://www.zeromq.org/intro:get-the-software) and " 
                      +"PYZMQ (http://zeromq.github.com/pyzmq/)")
        else:
            logger.error("No suitable coordination backend found.")
        
        logger.debug("Parsing URL: " + coordination_url)
        scheme, username, password, host, port, dbtype  = self.__parse_url(coordination_url) 
        
        if port == -1:
            port = None
        coordination = bigjob_coordination(server=host, server_port=port, username=username, 
                                           password=password, dbtype=dbtype, url_prefix=scheme)
        return coordination
    
   
            
    def start_pilot_job(self, 
                 lrms_url, 
                 bigjob_agent_executable=None,
                 number_nodes=1,
                 queue=None,
                 project=None,
                 working_directory=None,
                 userproxy=None,
                 walltime=None,
                 processes_per_node=1,
                 filetransfers=None):
        """ Start a batch job (using SAGA Job API) at resource manager. Currently, the following resource manager are supported:
            fork://localhost/ (Default Job Adaptor
            gram://qb1.loni.org/jobmanager-pbs (Globus Adaptor)
            pbspro://localhost (PBS Prop Adaptor)
        
        """
         
        if self.job != None:
            raise BigJobError("One BigJob already active. Please stop BigJob first.") 
            return

        ##############################################################################
        # initialization of coordination and communication subsystem
        # Communication & Coordination initialization
        lrms_saga_url = saga.url(lrms_url)
        self.url = lrms_saga_url
        self.pilot_url = self.app_url + ":" + lrms_saga_url.host
        pilot_url_dict[self.pilot_url]=self
        
        logger.debug("create pilot job entry on backend server: " + self.pilot_url)
        self.coordination.set_pilot_state(self.pilot_url, str(Unknown), False)
        self.coordination.set_pilot_description(self.pilot_url, filetransfers)    
        logger.debug("set pilot state to: " + str(Unknown))
        ##############################################################################
        
        self.number_nodes=int(number_nodes)        
        
        # create job description
        jd = saga.job.description()
        
        #  Attempt to create working directory (e.g. in local scenario)
        if working_directory != None:
            if not os.path.isdir(working_directory) \
                and (lrms_saga_url.scheme.startswith("fork") or lrms_saga_url.scheme.startswith("condor")) \
                and working_directory.startswith("go:")==False:
                os.mkdir(working_directory)
            self.working_directory = working_directory
        else:
            # if no working dir is set assume use home directory
            # will fail if home directory is not the same on remote machine
            # but this is just a guess to avoid failing
            self.working_directory = os.path.expanduser("~") 
            
            
        # Determine whether target machine use gsissh or ssh to logon.
        self.launch_method = self.__get_launch_method(lrms_saga_url.host,lrms_saga_url.username)
            
        ##############################################################################
        # File Management and Stage-In
        if lrms_saga_url.scheme.startswith("condor")==False:           
            # build target url for working directory
            # this will also create the remote directory for the BJ
            # Fallback if working directory is not a valid URL
            if not (self.working_directory.startswith("go:") or self.working_directory.startswith("ssh://")):            
                if lrms_saga_url.username!=None and lrms_saga_url.username!="":
                    bigjob_working_directory_url = "ssh://" + lrms_saga_url.username + "@" + lrms_saga_url.host + self.__get_bigjob_working_dir()
                else:
                    bigjob_working_directory_url = "ssh://" + lrms_saga_url.host + self.__get_bigjob_working_dir()
            else:
                # working directory is a valid file staging URL
                bigjob_working_directory_url=self.working_directory            
                
            # initialize file manager that takes care of file movement and directory creation
            if self.__filemanager==None:
                self.__initialize_pilot_data(bigjob_working_directory_url) # determines the url
            
            if self.__filemanager != None and not self.working_directory.startswith("/"):
                self.working_directory = self.__filemanager.get_path(bigjob_working_directory_url)
            
            # determine working directory of bigjob 
            # if a remote sandbox can be created via ssh => create a own dir for each bj job id
            # otherwise use specified working directory
            logger.debug("BigJob working directory: %s"%bigjob_working_directory_url)
            if self.__filemanager!=None and self.__filemanager.create_remote_directory(bigjob_working_directory_url)==True:
                self.working_directory = self.__get_bigjob_working_dir()
                self.__stage_files(filetransfers, bigjob_working_directory_url)
            else:        
                logger.warn("For file staging. SSH (incl. password-less authentication or Globus Online is required.")
            
            logger.debug("BJ Working Directory: %s", self.working_directory)      
        
        
        ##############################################################################
        # Create and process BJ bootstrap script
        logger.debug("Adaptor specific modifications: "  + str(lrms_saga_url.scheme))
        bootstrap_script = self.generate_bootstrap_script(self.coordination.get_address(), self.pilot_url)
        if lrms_saga_url.scheme == "gram":
            bootstrap_script = self.escape_rsl(bootstrap_script)
        elif lrms_saga_url.scheme == "pbspro" or lrms_saga_url.scheme=="xt5torque" or lrms_saga_url.scheme=="torque":                
            bootstrap_script = self.escape_pbs(bootstrap_script)
        elif lrms_saga_url.scheme == "ssh":
            bootstrap_script = self.escape_ssh(bootstrap_script)
        ############ submit pbs script which launches bigjob agent using ssh adaptors########## 
        elif lrms_saga_url.scheme == "pbs-ssh":
            bootstrap_script = self.escape_ssh(bootstrap_script)
            # PBS specific BJ plugin
            pbssshj = pbsssh(bootstrap_script, self.launch_method, queue, project,lrms_saga_url, walltime, number_nodes, 
                             processes_per_node, userproxy, self.working_directory, self.working_directory)
            self.job = pbssshj
            self.job.run()
            return
        ############ submit sge script which launches bigjob agent using ssh adaptors########## 
        elif lrms_saga_url.scheme == "sge-ssh":
            bootstrap_script = self.escape_ssh(bootstrap_script)
            # PBS specific BJ plugin
            sgesshj = sgessh(bootstrap_script, lrms_saga_url, walltime, number_nodes, 
                             processes_per_node, userproxy, project, queue, self.working_directory, self.working_directory)
            self.job = sgesshj
            self.job.run()
            return
        elif is_bliss:
            bootstrap_script = self.escape_bliss(bootstrap_script)

        # Define Agent Executable in Job description
        # in Condor case bootstrap script is staged 
        # (Python app cannot be passed inline in Condor job description)
        if lrms_saga_url.scheme.startswith("condor")==True:
            logger.debug("Using Condor")
            condor_bootstrap_filename = os.path.join("/tmp", "bootstrap-"+str(self.uuid))
            condor_bootstrap_file = open(condor_bootstrap_filename, "w")
            condor_bootstrap_file.write(bootstrap_script)
            condor_bootstrap_file.close()
           
            jd.executable = "/usr/bin/env"
            jd.arguments = ["python",  condor_bootstrap_filename]                
            bj_file_transfers = []
            file_transfer_spec = condor_bootstrap_filename + " > " + os.path.basename(condor_bootstrap_filename)
            bj_file_transfers.append(file_transfer_spec)
            output_file_transfer_spec = os.path.join(self.working_directory, "output.tar.gz") +" < output.tar.gz"
            bj_file_transfers.append(output_file_transfer_spec)
            if filetransfers != None:
                for t in filetransfers:
                    bj_file_transfers.append(t)
            logger.debug("Condor file transfers: " + str(bj_file_transfers))
            jd.file_transfer = bj_file_transfers
        else:
            if is_bliss==False:
                jd.number_of_processes = str(number_nodes)
                jd.processes_per_host=str(processes_per_node)
            else:
                jd.TotalCPUCount=str(int(number_nodes)*int(processes_per_node))                    
            jd.spmd_variation = "single"
            jd.arguments = ["python", "-c", bootstrap_script]
            jd.executable = "/usr/bin/env"           
       
        if queue != None:
            jd.queue = queue
        if project !=None:
            jd.job_project = [project]
        if walltime!=None:
            jd.wall_time_limit=str(walltime)
    
        if lrms_saga_url.scheme.startswith("condor")==False:
            jd.working_directory = self.working_directory
        else:
            jd.working_directory=""

        logger.debug("Working directory: " + jd.working_directory)
        
        #if lrms_saga_url.scheme.startswith("condor"):
        #    jd.output = "stdout-bigjob_agent.txt"
        #    jd.error = "stderr-bigjob_agent.txt"         
        #else:
        jd.output = os.path.join(self.working_directory, "stdout-bigjob_agent.txt")
        jd.error = os.path.join(self.working_directory, "stderr-bigjob_agent.txt")
          
           
        # Submit job
        js = None    
        if userproxy != None and userproxy != '':
            s = saga.session()
            os.environ["X509_USER_PROXY"]=userproxy
            ctx = saga.context("x509")
            ctx.set_attribute ("UserProxy", userproxy)
            s.add_context(ctx)
            logger.debug("use proxy: " + userproxy)
            js = saga.job.service(s, lrms_saga_url)
        else:
            logger.debug("use standard proxy")
            js = saga.job.service(lrms_saga_url)

        logger.debug("Creating pilot job with description: %s" % str(jd))
              

        self.job = js.create_job(jd)
        logger.debug("Submit pilot job to: " + str(lrms_saga_url))
        self.job.run()
        return self.pilot_url
        
    def generate_bootstrap_script(self, coordination_host, coordination_namespace):
        script = textwrap.dedent("""import sys
import os
import urllib
import sys
import time
start_time = time.time()
home = os.environ.get("HOME")
#print "Home: " + home
if home==None: home = os.getcwd()
BIGJOB_AGENT_DIR= os.path.join(home, ".bigjob")
if not os.path.exists(BIGJOB_AGENT_DIR): os.mkdir (BIGJOB_AGENT_DIR)
BIGJOB_PYTHON_DIR=BIGJOB_AGENT_DIR+"/python/"
if not os.path.exists(BIGJOB_PYTHON_DIR): os.mkdir(BIGJOB_PYTHON_DIR)
BOOTSTRAP_URL="https://raw.github.com/saga-project/BigJob/master/bootstrap/bigjob-bootstrap.py"
BOOTSTRAP_FILE=BIGJOB_AGENT_DIR+"/bigjob-bootstrap.py"
#ensure that BJ in .bigjob is upfront in sys.path
sys.path.insert(0, os.getcwd() + "/../")
#sys.path.insert(0, /User/luckow/.bigjob/python/lib")
#sys.path.insert(0, os.getcwd() + "/../../")
p = list()
for i in sys.path:
    if i.find(\".bigjob/python\")>1:
          p.insert(0, i)
for i in p: sys.path.insert(0, i)
print "Python path: " + str(sys.path)
print "Python version: " + str(sys.version_info)
try: import saga
except: print "SAGA and SAGA Python Bindings not found: BigJob only work w/ non-SAGA backends e.g. Redis, ZMQ.";
try: import bigjob.bigjob_agent
except: 
    print "BigJob not installed. Attempt to install it."; 
    opener = urllib.FancyURLopener({}); 
    opener.retrieve(BOOTSTRAP_URL, BOOTSTRAP_FILE); 
    print "Execute: " + "python " + BOOTSTRAP_FILE + " " + BIGJOB_PYTHON_DIR
    os.system("/usr/bin/env")
    try:
        os.system("python " + BOOTSTRAP_FILE + " " + BIGJOB_PYTHON_DIR); 
        activate_this = BIGJOB_PYTHON_DIR+'bin/activate_this.py'; 
        execfile(activate_this, dict(__file__=activate_this))
    except:
        print "BJ installation failed. Trying system-level python (/usr/bin/python)";
        os.system("/usr/bin/python " + BOOTSTRAP_FILE + " " + BIGJOB_PYTHON_DIR); 
        activate_this = BIGJOB_PYTHON_DIR+'bin/activate_this.py'; 
        execfile(activate_this, dict(__file__=activate_this))
#try to import BJ once again
import bigjob.bigjob_agent
# execute bj agent
args = list()
args.append("bigjob_agent.py")
args.append(\"%s\")
args.append(\"%s\")
print "Bootstrap time: " + str(time.time()-start_time)
print "Starting BigJob Agents with following args: " + str(args)
bigjob_agent = bigjob.bigjob_agent.bigjob_agent(args)
""" % (coordination_host, coordination_namespace))
        return script
    
    def escape_rsl(self, bootstrap_script):
        logger.debug("Escape RSL")
        bootstrap_script = bootstrap_script.replace("\"", "\"\"")        
        return bootstrap_script
            
    def escape_pbs(self, bootstrap_script):
        logger.debug("Escape PBS")
        bootstrap_script = "\'" + bootstrap_script+ "\'"
        return bootstrap_script
    
    
    def escape_ssh(self, bootstrap_script):
        logger.debug("Escape SSH")
        bootstrap_script = bootstrap_script.replace("\"", "\\\"")
        bootstrap_script = bootstrap_script.replace("\'", "\\\"")
        bootstrap_script = "\"" + bootstrap_script+ "\""
        return bootstrap_script
    
    def escape_bliss(self, bootstrap_script):
        logger.debug("Escape fork")
        bootstrap_script = bootstrap_script.replace("\'", "\"")
        bootstrap_script = "\'" + bootstrap_script+ "\'"
        return bootstrap_script
     
    def list_subjobs(self):
        sj_list = self.coordination.get_jobs_of_pilot(self.pilot_url)
        logger.debug(str(sj_list))
        subjobs = []
        for i in sj_list:
            url = i 
            if url.find("/")>0:
                url = url[url.find("bigjob"):]
                url =  url.replace("/", ":")    
            sj = subjob(job_url=url)
            subjobs.append(sj)
        return subjobs
     
    def add_subjob(self, jd, job_url, job_id):
        if jd.attribute_exists ("FileTransfer"):
            try:
                logger.debug("Stage input files for sub-job")
                self.__stage_files(jd.file_transfer, self.__get_subjob_working_dir(job_id))
            except:
                logger.error("File Stagein failed. Is Paramiko installed?")
        logger.debug("add subjob to queue of PJ: " + str(self.pilot_url))        
        for i in range(0,3):
            try:
                logger.debug("create dictionary for job description. Job-URL: " + job_url)
                # put job description attributes to Redis
                job_dict = {}
                #to accomendate current bug in bliss (Number of processes is not returned from list attributes)
                job_dict["NumberOfProcesses"] = "1" 
                attributes = jd.list_attributes()   
                logger.debug("SJ Attributes: " + str(jd))             
                for i in attributes:          
                        if jd.attribute_is_vector(i):
                            #logger.debug("Add attribute: " + str(i) + " Value: " + str(jd.get_vector_attribute(i)))
                            vector_attr = []
                            for j in jd.get_vector_attribute(i):
                                vector_attr.append(j)
                            job_dict[i]=vector_attr
                        else:
                            #logger.debug("Add attribute: " + str(i) + " Value: " + jd.get_attribute(i))
                            job_dict[i] = jd.get_attribute(i)
                
                job_dict["state"] = str(Unknown)
                job_dict["job-id"] = str(job_id)
                
                #logger.debug("update job description at communication & coordination sub-system")
                self.coordination.set_job(job_url, job_dict)                                                
                self.coordination.queue_job(self.pilot_url, job_url)
                break
            except:
                traceback.print_exc(file=sys.stdout)
                time.sleep(2)
                #raise Exception("Unable to submit job")
                     
    def delete_subjob(self, job_url):
        self.coordination.delete_job(job_url) 
    
    def get_subjob_state(self, job_url):
        return self.coordination.get_job_state(job_url) 
    
    def get_subjob_details(self, job_url):
        return self.coordination.get_job(job_url) 
     
    def get_state(self):        
        """ duck typing for get_state of saga.job.job  
            state of saga job that is used to spawn the pilot agent
        """
        try:
            return self.job.get_state()
        except:
            return self.get_state_detail()
            #return None
    
    def get_state_detail(self): 
        """ internal state of BigJob agent """ 
        try:
            return self.coordination.get_pilot_state(self.pilot_url)["state"]
        except:
            return None
    
    def get_free_nodes(self):
        jobs = self.coordination.get_jobs_of_pilot(self.pilot_url)
        number_used_nodes=0
        for i in jobs:
            job_detail = self.coordination.get_job(i)            
            if job_detail != None and job_detail.has_key("state") == True\
                and job_detail["state"]==str(Running):
                job_np = "1"
                if (job_detail["NumberOfProcesses"] == True):
                    job_np = job_detail["NumberOfProcesses"]
                number_used_nodes=number_used_nodes + int(job_np)
        return (self.number_nodes - number_used_nodes)

    
    def stop_pilot_job(self):
        """ mark in advert directory of pilot-job as stopped """
        try:
            logger.debug("stop pilot job: " + self.pilot_url)
            self.coordination.set_pilot_state(self.pilot_url, str(Done), True)            
            self.job=None
        except:
            pass
    
    def cancel(self):        
        """ duck typing for cancel of saga.cpr.job and saga.job.job  """
        logger.debug("Cancel Pilot Job")
        try:
            if self.url.scheme.startswith("condor")==False:
                self.job.cancel()
            else:
                print "Output files are being transfered to file: outpt.tar.gz. Please wait until transfer is complete."
        except:
            pass
            #traceback.print_stack()
        try:            
            self.stop_pilot_job()
            logger.debug("delete pilot job: " + str(self.pilot_url))                      
            if CLEANUP:
                self.coordination.delete_pilot(self.pilot_url)                    
            #os.remove(os.path.join("/tmp", "bootstrap-"+str(self.uuid)))            
        except:
            pass
            #traceback.print_stack()

    def wait(self):
        """ Waits for completion of all sub-jobs """        
        while 1:
            jobs = self.coordination.get_jobs_of_pilot(self.pilot_url)
            finish_counter=0
            result_map = {}
            for i in jobs:
                state = self.coordination.get_job_state(str(i))            
                #state = job_detail["state"]                
                if result_map.has_key(state)==False:
                    result_map[state]=1
                else:
                    result_map[state] = result_map[state]+1
                if self.__has_finished(state)==True:
                    finish_counter = finish_counter + 1                   
            logger.debug("Total Jobs: %s States: %s"%(len(jobs), str(result_map)))
            if finish_counter == len(jobs):
                break
            time.sleep(2)


    ###########################################################################
    # internal methods
    
    def __has_finished(self, state):
        state = state.lower()
        if state=="done" or state=="failed" or state=="canceled":
            return True
        else:
            return False
    
    def __parse_url(self, url):
        try:
            if is_bliss==True:
                raise BigJobError("BLISS URL broken.")
            surl = saga.url(url)
            host = surl.host
            port = surl.port
            username = surl.username
            password = surl.password
            query = surl.query
            scheme = "%s://"%surl.scheme
        except:
            """ Fallback URL parser based on Python urlparse library """
            logger.error("URL %s could not be parsed"%(url))
            traceback.print_exc(file=sys.stderr)
            result = urlparse.urlparse(url)
            logger.debug("Result: " + str(result))
            host = result.hostname
            #host = None
            port = result.port
            username = result.username
            password = result.password
            scheme = "%s://"%result.scheme 
            if host==None:
                logger.debug("Python 2.6 fallback")
                if url.find("/", len(scheme)) > 0:
                    host = url[len(scheme):url.find("/", len(scheme))]
                else:
                    host = url[len(scheme):]
                if host.find(":")>1:
                    logger.debug(host)
                    comp = host.split(":")
                    host = comp[0]
                    port = int(comp[1])
                    
            if url.find("?")>0:
                query = url[url.find("?")+1:]
            else:
                query = None
            
        
        logger.debug("%s %s %s"%(scheme, host, port))
        return scheme, username, password, host, port, query     
            
    
    def __get_bigjob_working_dir(self):
        if self.working_directory.find(self.uuid)!=-1: # working directory already contains BJ id
            return self.working_directory
        else:
            return os.path.join(self.working_directory, self.uuid)
    
    
    def __get_subjob_working_dir(self, sj_id):
        return os.path.join(self.__get_bigjob_working_dir(), sj_id)
    

    ###########################################################################
    # File Management
    
    def __initialize_pilot_data(self, service_url):
        # initialize file adaptor
        # Pilot Data API for File Management
        if service_url.startswith("ssh:"):
            logger.debug("Use SSH backend")
            try:
                from pilot.filemanagement.ssh_adaptor import SSHFileAdaptor
                self.__filemanager = SSHFileAdaptor(service_url) 
            except:
                logger.warn("SSH/Paramiko package not found.")            
        elif service_url.startswith("http:"):
            logger.debug("Use WebHDFS backend")
            try:
                from pilot.filemanagement.webhdfs_adaptor import WebHDFSFileAdaptor
                self.__filemanager = WebHDFSFileAdaptor(service_url)
            except:
                logger.warn("WebHDFS package not found.")        
        elif service_url.startswith("go:"):
            logger.debug("Use Globus Online backend")
            try:
                from pilot.filemanagement.globusonline_adaptor import GlobusOnlineFileAdaptor
                self.__filemanager = GlobusOnlineFileAdaptor(service_url)
            except:
                logger.warn("Globus Online package not found.") 
                self.__print_traceback()
                
            
                  

    def __stage_files(self, filetransfers, target_url):
        logger.debug("Stage: %s to %s"%(filetransfers, target_url))
        if filetransfers==None:
            return       
        self.__filemanager.create_remote_directory(target_url)
        for i in filetransfers:
            source_file=i
            if i.find(">")>0:
                source_file = i[:i.find(">")].strip()
            if source_file.startswith("ssh://")==False and source_file.startswith("go://")==False:
                logger.error("Staging of file: %s not supported. Please use URL in form ssh://<filename>"%source_file)
                continue
            target_url_full = os.path.join(target_url, os.path.basename(source_file))
            logger.debug("Stage: %s to %s"%(source_file, target_url_full))
            #self.__third_party_transfer(source_file, target_url_full)
            self.__filemanager.transfer(source_file, target_url_full)
           
       

    def __get_launch_method(self, hostname, user=None):
        """ returns desired execution method: ssh, aprun """
        if user == None: user = self.__discover_ssh_user(hostname)
        if user!=None: logger.debug("discovered user: " + user)
        gsissh_available = False
        try:
            gsissh_available = (subprocess.call("gsissh "+ user + "@" + hostname+" /bin/date", shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE)==0)
        except:
            pass
    
        ssh_available = False
        try:
            ssh_available = (subprocess.call("ssh "+ user + "@" + hostname+" /bin/date", shell=True, stdout=subprocess.PIPE,stderr=subprocess.PIPE)==0)
        except:
            pass
    
        launch_method = "ssh"
        if ssh_available == False and gsissh_available == True:
            launch_method="gsissh"
        else:
            launch_method="ssh"
        logging.debug(" launch method used " + str(launch_method))
        return launch_method
    
    
    def __discover_ssh_user(self, hostname):
        # discover username
        user = None
        ssh_config = os.path.join(os.path.expanduser("~"), ".ssh/config")
        ssh_config_file = open(ssh_config, "r")
        lines = ssh_config_file.readlines()
        for i in range(0, len(lines)):
            line = lines[i]
            if line.find(hostname)>0:
                for k in range(i + 1, len(lines)):
                    sub_line = lines[k]
                    if sub_line.startswith(" ")==True and sub_line.startswith("\t")==True:
                        break # configuration for next host
                    elif sub_line.find("User")!=-1:
                        stripped_sub_line = sub_line.strip()
                        user = stripped_sub_line.split()[1]
                        break
        ssh_config_file.close() 
        return user
    
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
        
    def __repr__(self):
        return self.pilot_url 

    def __del__(self):
        """ BJ is not cancelled when object terminates
            Application can reconnect to BJ via pilot url later on"""
        pass
        #self.cancel()


                    
                    
class subjob(api.base.subjob):
    
    def __init__(self, coordination_url=None, job_url=None):
        """Constructor"""
        
        self.coordination_url = coordination_url
        if job_url!=None:
            self.job_url=job_url
            self.uuid = self.__get_sj_id(job_url)
            self.pilot_url = self.__get_pilot_url(job_url)
            logger.debug("Reconnect SJ: %s Pilot %s"%(self.job_url, self.pilot_url))
        else:
            self.uuid = "sj-" + str(get_uuid())
            self.job_url = None
            self.pilot_url = None
            self.bj = None
            
    
    def __get_sj_id(self, job_url):
        start = job_url.index("sj-")        
        return job_url[start:]
    
    
    def __get_pilot_url(self, job_url):
        end =job_url.index(":jobs")
        return job_url[:end]    
    
                    
    def get_job_url(self, pilot_url):
        self.job_url = pilot_url + ":jobs:" + str(self.uuid)
        return self.job_url
    

    def submit_job(self, pilot_url, jd):
        """ submit subjob to referenced bigjob """
        if self.job_url==None:
            self.job_url=self.get_job_url(pilot_url)            
        
        if self.pilot_url==None:
            self.pilot_url = pilot_url
            self.bj=pilot_url_dict[pilot_url]    
        self.bj.add_subjob(jd, self.job_url, self.uuid)


    def get_state(self, pilot_url=None):        
        """ duck typing for saga.job  """
        if self.pilot_url==None:
            self.pilot_url = pilot_url
            self.bj=pilot_url_dict[pilot_url]                
        return self.bj.get_subjob_state(self.job_url)
    
    
    def cancel(self, pilot_url=None):
        logger.debug("delete job: " + self.job_url)
        if self.pilot_url==None:
            self.pilot_url = pilot_url
            self.bj=pilot_url_dict[pilot_url]  
        if str(self.bj.get_state())=="Running":
            self.bj.delete_subjob(self.job_url)        
        
    def get_exe(self, pilot_url=None):
        if self.pilot_url==None:
            self.pilot_url = pilot_url
            self.bj=pilot_url_dict[pilot_url]  
        sj = self.bj.get_subjob_details(self.job_url)
        return sj["Executable"]
   
    def get_arguments(self, pilot_url=None):
        if self.pilot_url==None:
            self.pilot_url = pilot_url
            self.bj=pilot_url_dict[pilot_url]  
        sj = self.bj.get_subjob_details(self.job_url)  
        #logger.debug("Subjob details: " + str(sj))              
        arguments=""
        for  i in  sj["Arguments"]:
            arguments = arguments + " " + i
        return arguments

    def __del__(self):
        """ Do nothing - keep subjobs alive and allow a later reconnection"""
        pass
        #self.cancel()
    
    def __repr__(self):        
        if(self.job_url==None):
            return "None"
        else:
            return self.job_url
        
        
class description(saga.job.description):
    pass



""" Example application demonstrating job submission via bigjob 

    DON'T EDIT THIS FILE (UNLESS THERE IS A BUG)
    
    THIS FILE SHOULD NOT BE COMMITTED TO SVN WITH USER-SPECIFIC PATHS!
"""
import os
import time
import pdb
import sys

# configuration
""" This variable defines the coordination system that is used by BigJob
    e.g. 
        advert://localhost (SAGA/Advert SQLITE)
        advert://advert.cct.lsu.edu:8080 (SAGA/Advert POSTGRESQL)
        advert://advert.cct.lsu.edu:5432 (SAGA/Advert POSTGRESQL)
        redis://localhost:6379 (Redis at localhost)
        tcp://localhost (ZMQ)
        tcp://* (ZMQ - listening to all interfaces)
"""

COORDINATION_URL = "advert://localhost/?dbtype=sqlite3"
#COORDINATION_URL = "advert://SAGA:SAGA_client@advert.cct.lsu.edu:8080/?dbtype=postgresql"
#COORDINATION_URL = "tcp://*"
#COORDINATION_URL = "redis://localhost:6379"
#COORDINATION_URL = "redis://gw68.quarry.iu.teragrid.org:2525"
#COORDINATION_URL="sqlasyncadvert://gw68.quarry.iu.teragrid.org/"

# for running BJ from local dir
sys.path.insert(0, os.getcwd() + "/../")

from bigjob import bigjob, subjob, description


def main():
    # Start BigJob

    ##########################################################################################
    # Edit parameters for BigJob
    queue=None # if None default queue is used
    project=None # if None default allocation is used 
    walltime=10
    processes_per_node=4
    number_of_processes = 8
    workingdirectory=os.path.join(os.getcwd(), "agent")  # working directory for agent
    userproxy = None # userproxy (not supported yet due to context issue w/ SAGA)

    
    """ 
    URL of the SAGA Job Service that is used to dispatch the pilot job.
    The following URLs are accepted:
    
    lrms_url = "gram://oliver1.loni.org/jobmanager-pbs" # globus resource url used when globus is used. (LONI)
    lrms_url = "pbspro://louie1.loni.org" # pbspro resource url used when pbspro scheduling system is used.(Futuregrid or LSU Machines)
    lrms_url = "ssh://louie1.loni.org" # ssh resource url which launches jobs on target machine. Jobs not submitted to scheduling system.
    lrms_url = "pbs-ssh://louie1.loni.org" # Submit jobs to scheduling system of remote machine.
    lrms_url = "xt5torque://localhost" # torque resource url.
    
    Please ensure that the respective SAGA adaptor is installed and working
    """
    lrms_url = "fork://localhost" # resource url to run the jobs on localhost
   
    ##########################################################################################

    print "Start Pilot Job/BigJob at: " + lrms_url
    bj = bigjob(COORDINATION_URL)
    bj.start_pilot_job( lrms_url,
                        None,
                        number_of_processes,
                        queue,
                        project,
                        workingdirectory,
                        userproxy,
                        walltime,
                        processes_per_node)
    
    print "Pilot Job/BigJob URL: " + bj.pilot_url + " State: " + str(bj.get_state())

    ##########################################################################################
    # Submit SubJob through BigJob
    jd = description()
    jd.executable = "/bin/date"
    jd.number_of_processes = "1"
    jd.spmd_variation = "single"
    jd.arguments = [""]
    #jd.working_directory = "/tmp" 
    jd.output = "stdout.txt"
    jd.error = "stderr.txt"
    sj = subjob()
    sj.submit_job(bj.pilot_url, jd)
    
    #########################################
    # busy wait for completion
    while 1:
        state = str(sj.get_state())
        print "state: " + state
        if(state=="Failed" or state=="Done"):
            break
        time.sleep(2)

    ##########################################################################################
    # Cleanup - stop BigJob
    bj.cancel()
    #time.sleep(30)


""" Test Job Submission via Advert """
if __name__ == "__main__":
    main()
    



#!/usr/bin/env python


#############################################################
# SAGA ManyJob 
#############################################################


print("############################################################")
print("Make sure you are running this example on head node")
print("############################################################")

# import required packages and check for their availability
try:
    import sys
    import getopt
    import saga
    import time
    import pdb
    import os
    import traceback
    import logging
    import sys
except:
    print("Error: Libraries are missing. Please check SAGA Installation, SAGA Python bindings and other imported packages")
    print(" Error :Please launch python; import saga; if it works then python bindings & saga installation is successful")
    sys.exit(0);

#############################################################
#  Set BigJob Path
#############################################################
# for running BJ from local dir
sys.path.insert(0, os.getcwd() + "/../")
try:
    from bigjob.bigjob_manager import *
    from bigjob_dynamic.many_job import *
except:
    print("Error: SAGA BigJob & ManyJob modules are not found")

#############################################################
# Set number of jobs
#############################################################

NUMBER_JOBS=8

#############################################################
# Grid certificate validation
#############################################################
grid_validity=os.system("globus-job-run localhost /bin/date > /dev/null 2>&1")
if grid_validity != 0:
    print(" Grid proxy not initated. Please use grid-proxy-init to initate grid proxy")
   

############################################################
# Functions used to determine state of jobs
############################################################

def has_finished(state):
        state = state.lower()
        if state=="done" or state=="failed" or state=="canceled":
            return True
        else:
            return False



if __name__ == "__main__":
    try:

        print " Number of jobs submitted " + str(NUMBER_JOBS) + " jobs."
        starttime=time.time()

        resource_list = []
        #resource_list.append( {"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_nodes" : "64", "allocation" : "<your allocation>", "queue" : "workq", "bigjob_agent": (os.getcwd() + "/bigjob_agent_launcher.sh") , "working_directory": (os.getcwd() + "/agent"), "walltime":10 })
        #resource_list.append({"resource_url" : "gram://oliver1.loni.org/jobmanager-pbs", "number_nodes" : "1", "processes_per_node":"4", "allocation" : None, "queue" : None, "bigjob_agent": (BIGJOB_HOME + "/bigjob_agent_launcher.sh"), "working_directory": (os.getcwd() + "/agent"), "walltime":30 }) 
        #resource_list.append({"resource_url" : "gram://eric1.loni.org/jobmanager-pbs", "number_nodes" : "2", "processes_per_node":"4", "allocation" : None, "queue" : "checkpt", "bigjob_agent": (BIGJOB_HOME + "/bigjob_agent_launcher.sh"), "working_directory": (os.getcwd() + "/agent"), "walltime":4320 }) 
       
        # resource_url - Machine on which you want to run the job.
        # number_of_processes - total number of processes requested on that machine.
        # processes_per_node
        # allocation - if None then default allocation is used.
        # queue - if None then default queue is used.
        
        resource_list.append({"resource_url" : "gram://eric1.loni.org/jobmanager-pbs", 
                              "number_of_processes" : "4", "processes_per_node":"4", 
                              "allocation" : None, "queue" : "checkpt", 
                              "working_directory": (os.getcwd() + "/agent"), 
                              "walltime":20 }) 

        resource_list.append({"resource_url" : "gram://eric1.loni.org/jobmanager-pbs", 
                              "number_of_processes" : "4", "processes_per_node":"4", 
                              "allocation" : None, "queue" : "checkpt", 
                              "working_directory": (os.getcwd() + "/agent"), 
                              "walltime":20 }) 


        print "Create manyjob service "
        mjs = many_job_service(resource_list, COORDINATION_URL)
        jobs = []
        job_start_times = {}
        job_states = {}
        job_ids={}
        cwd = os.getcwd()

        
        for i in range(0, NUMBER_JOBS):
            jd = saga.job.description()
            jd.executable = "/bin/date"
            jd.number_of_processes = "1"
            jd.spmd_variation = "single"
            jd.arguments = [""]
            jd.working_directory = os.getcwd()
            jd.output =  "stdout-" + str(i) + ".txt"
            jd.error = "stderr-" + str(i) + ".txt"
            subjob = mjs.create_job(jd)
            subjob.run()
            print "Submited sub-job " + str(i) + "."
            jobs.append(subjob)
            job_start_times[subjob]=time.time()
            job_states[subjob] = subjob.get_state()


        print "************************ All Jobs submitted ************************"
        while 1: 
            finish_counter=0
            result_map = {}
            print("JOB_ID     Status           Machine       Job_Attributes")
            print("==========================================================")
            for i in range(0, NUMBER_JOBS):
                old_state = job_states[jobs[i]]
                state = jobs[i].get_state()
                if result_map.has_key(state) == False:
                    result_map[state]=0
                result_map[state] = result_map[state]+1
                k = str(jobs[i])                
                if k is not "None":                    
                    jobid = (k.split(":"))[4]
                    machine = (k.split(":"))[2] 
                    exe = jobs[i].get_exe() 
                    arguments = jobs[i].get_arguments()
                    strings = str(jobid) + "   " +  "%7s     %10s %10s,%6s" % (str(state),str(machine),str(exe), str(arguments))
                    print(strings)
                ##print "counter: " + str(i) + " job: " + str(jobs[i]) + " state: " + state
                #if old_state != state:
                    #print "Job " + str(jobs[i]) + " changed from: " + old_state + " to " + state
                #if old_state != state and has_finished(state)==True:
                    #print "Job: " + str(jobs[i]) + " Runtime: " + str(time.time()-job_start_times[jobs[i]]) + " s."
                if has_finished(state)==True:
                    finish_counter = finish_counter + 1
                job_states[jobs[i]]=state
                
            print "Current states: " + str(result_map) 
            time.sleep(5)
            if finish_counter == NUMBER_JOBS:
                break

        print(" All Jobs completed. Cleaning up the bigjob..." )
        mjs.cancel()
        runtime = time.time()-starttime
        print "Runtime: " + str(runtime) + " s; Runtime per Job: " + str(runtime/NUMBER_JOBS)
    except:
        traceback.print_exc(file=sys.stdout)
        try:
            k=0
            mjs.cancel()
        except:
            pass




#!/usr/bin/env python

"""Dynamic BigJob (ManyJob): Manages multiple BigJob (on different resources). 
Sub-jobs are distributed across the set of BJs managed by the dynamic BJ.
"""
import pdb
import sys
import os
sys.path.append(os.path.dirname( __file__ ))
import getopt
import saga
import time
import uuid
import socket
import traceback

import Queue
import threading
import logging
import time
import math
import operator
import copy


import bigjob.bigjob_manager 

# Log everything, and send it to stderr.
logging.basicConfig(level=logging.DEBUG)

COORDINATION_URL="advert://advert.cct.lsu.edu:8080"

class many_job_service(object):

    def __init__(self, bigjob_list, coordination_url):
        """ accepts resource list as key/value pair:
            ( {"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_nodes" : "32", "allocation" : "loni_stopgap2", "queue" : "workq", "bigjob_agent": "$(HOME)/src/REMDgManager/bigjob/advert_launcher.sh",  "walltime":1000},
              {"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_nodes" : "32", "allocation" : "loni_stopgap2", "queue" : "workq", "bigjob_agent": "$(HOME)/src/REMDgManager/bigjob/advert_launcher.sh", "walltime":1000})
        """        
        self.uuid = uuid.uuid1()
        
        if coordination_url==None:
            self.advert_host=COORDINATION_URL   
        else:
            self.advert_host=coordination_url

        # list of resource dicts (1 dict per resource) 
        # will also store state of bigjob
        self.bigjob_list=copy.deepcopy(bigjob_list) 
       
        # state variable storing state of sub-jobs 
        self.active_subjob_list = []
        self.subjob_bigjob_dict = {}

        # queue contains unscheduled subjobs        
        self.subjob_queue = Queue.Queue()
        
        # submit bigjobs to resources
        self.__init_bigjobs()
        
        # thread which tries to resubmit jobs
        self.stop=threading.Event()
        self.rescheduler_thread=threading.Thread(target=self.__reschedule_subjobs_thread)
        self.rescheduler_thread.start()
        
        # last queue Size
        self.last_queue_size = 0
        self.submisssion_times=[]

    def __init_bigjobs(self):
        """ start on specified resources a bigjob """
        self.bigjob_list = self.__schedule_bigjobs()
        for i in self.bigjob_list:
            self.__start_bigjob(i)


    def __start_bigjob(self, bj_dict):
        """ private method - starts a bigjob on the defined resource """
        gram_url = bj_dict["resource_url"]
        logging.debug("start bigjob at: " + gram_url)
        bj = bigjob.bigjob_manager.bigjob(self.advert_host)
        ppn="1"
        if ("processes_per_node" in bj_dict):
            ppn=bj_dict["processes_per_node"]
        else:
            bj_dict["processes_per_node"]="1"

        walltime = 3600
        if ("walltime" in bj_dict):
            walltime=bj_dict["walltime"]

        working_directory = (os.getcwd()+"/agent")
        if ("working_directory" in bj_dict):
            working_directory=bj_dict["working_directory"]

        bj.start_pilot_job(gram_url,
                           None,
                           bj_dict["number_of_processes"],
                           bj_dict["queue"],
                           bj_dict["allocation"],
                           working_directory, 
                           None,
                           walltime,
                           ppn)
        bj_dict["bigjob"]=bj # store bigjob for later reference in dict
        bj_dict["free_cores"]=int(bj_dict["number_of_processes"])
        bj_dict["to_be_terminated"]=False
        # lock for modifying the number of free nodes
        bj_dict["lock"] = threading.Lock()

    def add_resource(self, resource_dictionary):
        """ adds bigjob described in resource_dictionary to resources """
        bj_dict = copy.deepcopy(resource_dictionary)
        
        self.__start_bigjob(bj_dict)
        self.bigjob_list.append(bj_dict)
        return bj_dict["bigjob"] 
        
    
    def remove_resource(self, bigjob):
        """ remove bigjob from resource list of manyjob """
        # mark bigjob for termination (after all sub-jobs in bj are 
        # finished        
        bigjob["to_be_terminated"]=True
        
 
    def __cleanup_resources(self):
        """ called periodically from scheduling thread
            terminates big-jobs which are marked and don't have 
            any running sub-jobs
        """
        # iterate over copy of list, but remove from orig list
        for i in self.bigjob_list[:]:
            if i["to_be_terminated"]==True:
                bj = i["bigjob"]
                total_cores = int(i["number_of_processes"])
                if  i["free_cores"]==total_cores and not i.has_key("bj_stopped"):
                    logging.debug("***Stop BigJob: " + str(bj.pilot_url))
                    # release resources of pilot job
                    bj.stop_pilot_job()
                    i["bj_stopped"]=True
                    #self.bigjob_list.remove(i)

    
    def get_resources(self):
        """ returns list with bigjob dictionaries
            for each managed bigjob 1 dictionary exists 
        """
        return self.bigjob_list


    def list_bigjobs(self):
        """ returns a list of bigjob objects """
        return [i["bigjob"]  for i in self.bigjob_list]

    def __schedule_bigjobs(self):
        """ prioritizes bigjob_list (bigjob with shortest expected delay will have index 0) """
        # no scheduling for now (start bigjob in the user specified order)
        return self.bigjob_list

    def create_job (self, job_description):
        subjob = sub_job(self, job_description, self.advert_host)        
        return subjob

    def __run_subjob(self, subjob):
        # select appropriate bigjob
        st = time.time()
        bigjob_info = self.__schedule_subjob(subjob)
        job = subjob.job
        if bigjob_info == None:
            return job

        # create subjob on bigjob
        bj = bigjob_info["bigjob"]
        
        job.submit_job(bj.pilot_url, subjob.job_description)
        self.submisssion_times.append(time.time()-st)

        # store reference of subjob for further bookkeeping    
        self.active_subjob_list.append(subjob)
        self.subjob_bigjob_dict[subjob] = bigjob_info
        logging.debug("Subjob submission time: " + str(time.time()-st) + " sec.")
        return job

    def queue_subjob(self, subjob):
        subjob.job = bigjob.bigjob_manager.subjob(self.advert_host) 
        self.subjob_queue.put(subjob)
        return subjob.job

    def __schedule_subjob (self, subjob):
        """ find resource (bigjob) for subjob
            returns bigjob object """
        for i in self.bigjob_list:
            bigjob = i["bigjob"]
            lock = i["lock"]
            lock.acquire()
            free_cores = i["free_cores"]
            bigjob_url = bigjob.pilot_url
            state = bigjob.get_state_detail()
            logging.debug("Big Job: " + bigjob_url + " Cores: " + "%s"%free_cores + "/" 
                          + str(int(i["number_of_processes"])) 
                          + " State: " + str(state) + " Terminated: " + str(i["to_be_terminated"])
                          + " #Required Cores: " + subjob.job_description.number_of_processes
                          )
            if (state.lower() == "running" and free_cores >= int(subjob.job_description.number_of_processes) 
                and i["to_be_terminated"]==False):
                logging.debug("FOUND match - dispatch to BigJob: " + bigjob_url)
                free_cores = i["free_cores"]
                free_cores = free_cores - int(subjob.job_description.number_of_processes)
                i["free_cores"]=free_cores
                lock.release()
                return i 

            lock.release()

        # no resource found
        self.subjob_queue.put(subjob)
        logging.debug("found no active resource for sub-job => (re-) queue it")
        return None        

    def __check_subjobs_states(self):    
        """iterate through all sub-jobs and check state"""
        for i in self.active_subjob_list:            
            try:
                #logging.debug("get job state")
                state = i.job.get_state()
                #logging.debug("check job state")
                if self.__has_finished(state) == True:
                    #logging.debug("free resources")
                    self.__free_resources(i)
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=2, file=sys.stderr)
        
    def __free_resources(self, subjob):
        """free resources taken by subjob"""
        if(self.subjob_bigjob_dict.has_key(subjob)):
            logging.debug("job: " + str(subjob) + " done - free resources")
            bigjob = self.subjob_bigjob_dict[subjob]
            lock = bigjob["lock"]
            lock.acquire()
            free_cores = bigjob["free_cores"]
            free_cores = free_cores + int(subjob.job_description.number_of_processes)
            bigjob["free_cores"]=free_cores
            del(self.subjob_bigjob_dict[subjob])
            lock.release()
            print "Freed resource - new state: Big Job: " +  bigjob["bigjob"].pilot_url + " Cores: " + "%s"%free_cores + "/" + str(int(bigjob["number_of_processes"])) 
    
    def __reschedule_subjobs_thread(self):
        """ periodically checks subjob_queue for unscheduled subjobs
            if a unscheduled job exists it is scheduled
        """

        while True and self.stop.isSet()==False:
            logging.debug("Reschedule Thread")
            # check sub-job state
            self.__check_subjobs_states()
            # remove unneeded big-jobs
            self.__cleanup_resources()
            subjob = self.subjob_queue.get()  
            # check whether this is a real subjob object  
            if isinstance(subjob, sub_job):
                self.__run_subjob(subjob)
                if self.last_queue_size == self.subjob_queue.qsize() or self.__get_total_free_cores()==0:
                    time.sleep(2) # sleep 30 s        

        logging.debug("Re-Scheduler terminated")


    def __get_free_cores(self, bigjob):
        """ return number of free cores if bigjob is active """
        #pdb.set_trace()
        if (bigjob["bigjob"].get_state_detail().lower()=="running" 
            and bigjob["to_be_terminated"]==False):
            return bigjob["free_cores"]

        return 0            

    def __get_total_free_cores(self):
        """ get's the total number of free cores from all active  bigjobs """
        free_cores = map(self.__get_free_cores, self.bigjob_list)
        #print "Free cores: " + str(free_cores)
        if len(free_cores)>0:
            total_free_cores = reduce(lambda x, y: x + y, free_cores)
            logging.debug("free_cores: " + str(free_cores) + " total_free_cores: " + str(total_free_cores))
            return total_free_cores
        return 0

    def cancel(self):
        logging.debug("Cancel re-scheduler thread")
        self.stop.set()
        # put object in queue to unlock the get() operation
        self.subjob_queue.put("dummy")
        self.rescheduler_thread.join()        
        logging.debug("Cancel many-job: kill all bigjobs")
        for i in self.bigjob_list:
            bigjob = i["bigjob"]
            bigjob.cancel()
        self.print_stats(self.submisssion_times, "Submission Times")

    def print_stats(self, times, description):
        try:
            n = len(times)
            sum = reduce(operator.add, times)
            mean = sum/n
            variance=0
            if n > 1:
                for i in times:
                    variance += (i - mean)**2
                variance /= (n-1)
                variance = math.sqrt(variance)
            print description + " Average: " + str(mean) + " Stdev: " + str(variance)
        except:
            pass
        
    def __has_finished(self, state):
        state = state.lower()
        if state=="done" or state=="failed" or state=="canceled":
            return True
        else:
            return False

    def __repr__(self):
        return str(self.uuid)

    def __del__(self):
        self.cancel()
        
        
                    

class sub_job(object):
    """ Class for controlling individual sub-jobs """
    
    def __init__(self, manyjob, job_description, advert_host):
        # store bigjob for reference
        self.manyjob=manyjob
        
        # init sub-job via advert
        self.job_description = job_description
        self.job = None

    def run(self):
        # select appropriate bigjob
        self.job = self.manyjob.queue_subjob(self)

    def get_state(self):     
        try:
            state = self.job.get_state()
            return state
        except:
            #traceback.print_stack()
            pass
        return "Unknown"

    def get_arguments(self):
        try:
            arguments = self.job.get_arguments()
            return arguments
        except:
            traceback.print_exc(file=sys.stdout)
            pass
        return ""

    def get_exe(self):
        try:
            exe  = self.job.get_exe()
            return exe
        except:
            traceback.print_exc(file=sys.stdout)
            pass
        return "Unknown"


    def cancel(self):
        return self.job.cancel()

    def wait(self):
        while 1:
            try:
                state = self.get_state()
                logging.debug("wait: state: " + state)
                if self.__has_finished(state) == True:
                    break
                time.sleep(2)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                pass
   
    def __del__(self):
        pass

    def __repr__(self):        
        return str(self.job)

""" Test Job Submission via ManyJob abstraction """
if __name__ == "__main__":
    try:
        print "Test ManyJob"
        # create job description
        jd = saga.job.description()
        jd.executable = "/bin/date"
        jd.number_of_processes = "1"
        jd.spmd_variation = "single"
        jd.arguments = [""]
        jd.working_directory = "/home/luckow"
        jd.output = "output.txt"
        jd.error = "error.txt"
        # submit via mj abstraction
        #resource_list =  ( {"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_nodes" : "128", "allocation" : "<your allocation>", "queue" : "workq", "bigjob_agent": "$(HOME)/src/REMDgManager/bigjob/advert_launcher.sh"},
        #                   {"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_nodes" : "64", "allocation" : "<your allocation>", "queue" : "workq", "bigjob_agent": "$(HOME)/src/REMDgManager/bigjob/advert_launcher.sh"})
        resource_list = []
        resource_list.append({"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_nodes" : "16", "allocation" : "<your allocation>", "queue" : "workq", "bigjob_agent": os.getcwd() + "/bigjob_agent_launcher.sh"})
        print "Create manyjob service " 
        mjs = many_job_service(resource_list, None)
        print "Create sub-job using manyjob " + str(mjs) 
        subjob = mjs.create_job(jd)
        print "Run sub-job"
        subjob.run()
        print "Wait for termination"
        subjob.wait()
        print "Sub-job state: " + str(subjob.get_state())
        mjs.cancel()
    except:
        try:
            if mjs != None:
                mjs.cancel()
        except:
            pass
